from datetime import UTC, datetime

TIME_ZERO = 2082841200  # = 1970-01-01 00:00:00


def convert_timestamp(t: int) -> datetime:
    """
    Convert the FontLab timestamp to a datetime object.
    """
    b = t.to_bytes(4, signed=True)
    t = int.from_bytes(b, signed=False)

    # 1970-01-01 is the earliest timestamp in FL notation, enforce it
    if t < TIME_ZERO:
        t = TIME_ZERO

    return datetime.fromtimestamp(t - TIME_ZERO, UTC)


def datetime_to_fl(d: datetime) -> int:
    ts = int(datetime.timestamp(d))
    ts += TIME_ZERO

    if ts < TIME_ZERO:
        ts = TIME_ZERO
    else:
        b = ts.to_bytes(4, signed=False)
        ts = int.from_bytes(b, signed=True)
    return ts



from typing import Any


def get_empty_glyph(num_masters: int) -> dict[str, Any]:
    return {
        "metrics": [(0, 0) for _ in range(num_masters)],
        "name": "",
        "nodes": [],
        "num_masters": num_masters,
    }



from vfbLib.parsers.base import BaseParser
from vfbLib.typing import CustomCmap


class CustomCmapParser(BaseParser):
    """
    A parser that reads the custom CMAP settings.
    """

    def _parse(self) -> list[CustomCmap]:
        count = self.read_value()
        values = []
        for _ in range(count):
            cmap = CustomCmap(
                language_id=self.read_value(),
                platform_id=self.read_value(),
                encoding_id=self.read_value(),
                format=self.read_value(),
                option=self.read_value(),
                page_name=self.read_str_with_len(),
            )
            values.append(cmap)
        return values



from typing import TYPE_CHECKING

from vfbLib.compilers.base import BaseCompiler

if TYPE_CHECKING:
    from vfbLib.typing import PCLTDict


class PcltCompiler(BaseCompiler):
    def _compile(self, data: "PCLTDict") -> None:
        self.write_value(data["font_number"], signed=False)
        for k in (
            "pitch",
            "x_height",
            "style",
            "type_family",
            "cap_height",
            "symbol_set",
        ):
            self.write_value(data[k])
        self.write_str(data["typeface"], 16)
        for value in data["character_complement"]:
            self.write_uint8(value)
        self.write_str(data["file_name"], 6)
        self.write_uint8(data["stroke_weight"])
        self.write_uint8(data["width_type"])
        self.write_uint8(data["serif_style"])



import logging

from vfbLib.compilers.base import BaseCompiler

logger = logging.getLogger(__name__)


class MMKernPairCompiler(BaseCompiler):
    def _compile(self, data: list[int]) -> None:
        for value in data:
            self.write_int16(value)



import codecs
from argparse import ArgumentParser
from difflib import HtmlDiff, unified_diff
from pathlib import Path

import orjson

from vfbLib.vfb.vfb import Vfb


def diffvfb():
    parser = ArgumentParser(description="vfbdiff\nCopyright (c) 2024 by LucasFonts")
    parser.add_argument(
        "--html",
        type=str,
        help="Output diff in HTML format to file path",
    )
    parser.add_argument(
        "file1",
        type=str,
        nargs=1,
        help="First input file path (.vfb)",
    )
    parser.add_argument(
        "file2",
        type=str,
        nargs=1,
        help="Second input file path (.vfb)",
    )
    args = parser.parse_args()
    if not args:
        parser.print_help()
        return

    vfb1_path = Path(args.file1[0])
    vfb2_path = Path(args.file2[0])

    vfb1 = Vfb(vfb_path=vfb1_path, timing=False)
    vfb2 = Vfb(vfb_path=vfb2_path, timing=False)
    vfb1.read()
    vfb2.read()
    vfb1.decompile()
    vfb2.decompile()
    vfb1_str = (
        orjson.dumps(
            vfb1.as_dict(),
            option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS | orjson.OPT_SORT_KEYS,
        )
        .decode()
        .splitlines()
    )
    vfb2_str = (
        orjson.dumps(
            vfb2.as_dict(),
            option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS | orjson.OPT_SORT_KEYS,
        )
        .decode()
        .splitlines()
    )
    if args.html:
        html_diff = HtmlDiff()
        html = html_diff.make_file(
            str(vfb1_str),
            str(vfb2_str),
            str(vfb1_path),
            str(vfb2_path),
            context=True,
            numlines=5,
        )
        with codecs.open(args.html, "wb", "utf-8") as f:
            f.write(html)
    else:
        d = unified_diff(vfb1_str, vfb2_str, str(vfb1_path), str(vfb2_path))
        for line in d:
            print(line)



from typing import TYPE_CHECKING

from vfbLib.compilers.base import BaseCompiler

if TYPE_CHECKING:
    from vfbLib.typing import CustomCmap


class CustomCmapCompiler(BaseCompiler):
    def _compile(self, data: "list[CustomCmap]") -> None:
        self.write_value(len(data))
        for cmap in data:
            for k in ("language_id", "platform_id", "encoding_id", "format", "option"):
                self.write_value(cmap[k])
            self.write_str_with_len(cmap["page_name"])



from __future__ import annotations

import logging
from math import atan2, degrees
from typing import TYPE_CHECKING

from vfbLib import DIRECTIONS
from vfbLib.parsers.base import BaseParser
from vfbLib.typing import GuideDict, GuidePropertiesDict, GuidePropertyDict
from vfbLib.value import read_value

if TYPE_CHECKING:
    from io import BytesIO

    from vfbLib.typing import MMGuidesDict


logger = logging.getLogger(__name__)


def parse_guides(stream: "BytesIO", num_masters: int, name: str) -> "MMGuidesDict":
    """
    Common parser for global and glyph guides. Used by `GlobalGuidesParser` and
    `GlyphParser`.

    Args:
        stream (BytesIO): The data stream.
        num_masters (int): The number of masters contained in the data.
        name (str): A name used in error messages to describe the current location, e.g.
            "glyph" or "global".

    Returns:
        MMGuidesDict: Information about horizontal and vertical multiple master guides.

    Raises:
        ValueError: If the data structure is incomplete.
    """
    guides: "MMGuidesDict" = {"h": [], "v": []}
    for direction in DIRECTIONS:
        num_guides = read_value(stream)
        if num_guides == 0:
            continue

        guides[direction] = [[] for _ in range(num_guides)]
        for i in range(num_guides):
            for _ in range(num_masters):
                try:
                    pos = read_value(stream)
                    angle = degrees(atan2(read_value(stream), 10000))
                    guides[direction][i].append(GuideDict(pos=pos, angle=angle))
                except ValueError:
                    logger.error(f"Missing {direction} guideline data ({name})")
                    raise

    return guides


class GlobalGuidesParser(BaseParser):
    def _parse(self) -> "MMGuidesDict":
        assert self.vfb is not None
        guides = parse_guides(self.stream, self.vfb.num_masters, "global")
        assert self.stream.read() == b""
        return guides


class GuidePropertiesParser(BaseParser):
    def _parse(self) -> GuidePropertiesDict:
        guides = GuidePropertiesDict(h=[], v=[])
        for direction in DIRECTIONS:
            while True:
                index = self.read_value()
                if index == 0:
                    break

                g = GuidePropertyDict(index=index)

                color_raw = self.read_value()
                if color_raw > -1:
                    color_bgr = "%06x" % color_raw
                    g["color"] = f"#{color_bgr[4:]}{color_bgr[2:4]}{color_bgr[:2]}"

                name = self.read_str_with_len()
                if name:
                    g["name"] = name

                guides[direction].append(g)

        return guides



import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vfbLib.typing import KerningClassFlagDict
    from vfbLib.ufo.typing import UfoGroups


logger = logging.getLogger(__name__)


def check_for_missing_glyphs(
    group_name: str,
    glyphs: list[str],
    glyphOrder: list[str],
    skip_missing_group_glyphs: bool = False,
) -> tuple[list[str], list[str]]:
    missing = [n for n in glyphs if n not in glyphOrder]
    num_missing = len(missing)
    num_glyphs = len(glyphs)
    if missing:
        if num_missing == num_glyphs:
            logger.warning(
                f"All glyphs in group '{group_name}' are missing from "
                f"the font: {', '.join(missing)}"
            )
        else:
            logger.warning(
                f"{num_missing} of {num_glyphs} glyphs in group "
                f"'{group_name}' are missing from the font: {', '.join(missing)}"
            )
        if skip_missing_group_glyphs:
            remaining_glyphs = list(set(glyphs) - set(missing))
            if not remaining_glyphs:
                logger.warning(f"Not adding empty group '{group_name}' to the UFO.")
                return glyphs, missing
            glyphs = [g for g in glyphs if g not in missing]
    return glyphs, missing


def transform_groups(
    orig_groups: "UfoGroups",
    kerning_class_flags: "KerningClassFlagDict",
    glyphOrder: list[str],
    skip_missing_group_glyphs: bool = False,
) -> "tuple[UfoGroups, list[str], dict[str, str]]":
    """
    Transform the raw groups data, a dict of group name to a list of glyph names, which
    may reference glyphs not present in the font.

    The glyph name lists in orig_groups must be sorted so that the "key glyph" (a
    concept which does exist in VFB but not in UFO) is the first glyph.

    Kerning groups are renamed by looking up the side flags in `kerning_class_flags` and
    adding the appropriate `public.kern...` prefix.

    Missing glyphs are removed from groups, and if a group becomes empty, the whole
    group is omitted from the result. This clean-up can be controlled via
    `skip_missing_group_glyphs`.

    Args:
        orig_groups (UfoGroups): The unprocessed groups.
        kerning_class_flags (KerningClassFlagDict): The dict containing information
            about which side of the kerning pair a group pertains to.
        glyphOrder (list[str]): The glyph order of the font. It is used to remove any
            glyphs that are not present in the font from the groups.
        skip_missing_group_glyphs (bool, optional): Whether do omit glyphs that are not
            present in the font from the resulting groups. Empty groups are omitted as
            well.Defaults to False.

    Returns:
        tuple[UfoGroups, list[str], dict[str, str]]: The groups for the UFO, the
            original group order, and a dict of group_name to key glyph name.
    """
    FIRST = 2**10
    SECOND = 2**11
    groups: "UfoGroups" = {}
    group_order = []
    key_glyphs: dict[str, str] = {}
    for name, glyphs in orig_groups.items():
        # Find key glyph(s)
        group_key_glyphs = [g.strip("'") for g in glyphs if g.endswith("'")]
        other_group_glyphs = [g for g in glyphs if not g.endswith("'")]

        is_kerning = name.startswith("_")
        num_keyglyphs = len(group_key_glyphs)
        if (is_kerning and num_keyglyphs != 1) or (
            not is_kerning and num_keyglyphs != 0
        ):
            # Kerning groups should have one key glyph
            # Other groups should have no key glyph
            logger.warning(
                f"Unexpected number of key glyphs in group {name}: {group_key_glyphs}"
            )
        glyphs = group_key_glyphs + other_group_glyphs
        num_glyphs = len(glyphs)

        # Check for missing glyphs
        glyphs, missing = check_for_missing_glyphs(
            name, glyphs, glyphOrder, skip_missing_group_glyphs
        )
        num_missing = len(missing)

        key_glyph = None
        for key_glyph_candidate in group_key_glyphs:
            if key_glyph_candidate in missing and num_missing != num_glyphs:
                # Only warn if not all group glyphs are missing from the font
                logger.warning(
                    f"Key glyph '{key_glyph}' for group '{name}' is missing from font."
                )
            else:
                key_glyph = key_glyph_candidate
                break

        # The final name in the UFO is different for kerning groups, which also must be
        # split per side of the kerning pair
        final_name = name
        if name.startswith("_"):
            # The group is a kerning group, build a valid UFO name in `final_name`
            if name in kerning_class_flags:
                flags = kerning_class_flags[name][0]
            else:
                # Don't add kerning groups without a side flag, as FLS ignores them when
                # compiling the kern feature
                logger.warning(f"Skipping kerning group without side flag: '{name}'")
                continue

            for side, sidename in ((FIRST, 1), (SECOND, 2)):
                if flags & side:
                    # Add a group for each side that applies
                    final_name = f"public.kern{sidename}.{name}"
                    if final_name in groups:
                        logger.warning(
                            f"Skipping duplicate kern{sidename} group: {final_name}"
                        )
                        continue

                    if final_name in key_glyphs:
                        if key_glyphs[final_name] != key_glyph:
                            logger.warning(
                                f"Ignoring duplicate group '{final_name}' with "
                                f"different key glyphs: {key_glyphs[final_name]} vs. "
                                f"{key_glyph}."
                            )
                    elif key_glyph is not None:
                        key_glyphs[final_name] = key_glyph
                    groups[final_name] = glyphs
                    group_order.append(final_name)
            continue

        # Non-kerning group, may still have a key glyph
        # (https://github.com/LucasFonts/vfbLib/issues/258)
        if final_name in key_glyphs:
            if key_glyphs[final_name] != key_glyph:
                logger.warning(
                    f"Ignoring duplicate group '{final_name}' with different "
                    f"key glyphs: {key_glyphs[final_name]} vs. {key_glyph}."
                )
        elif key_glyph is not None:
            key_glyphs[final_name] = key_glyph
        groups[final_name] = glyphs
        group_order.append(final_name)

    return groups, group_order, key_glyphs


def build_glyph_to_group_maps(
    groups: "UfoGroups",
) -> "tuple[UfoGroups, dict[str, str], dict[str, str]]":
    glyph_group_1: dict[str, str] = {}
    glyph_group_2: dict[str, str] = {}
    for group in groups.items():
        name, glyphs = group
        if name.startswith("public.kern1"):
            for g in glyphs:
                if g in glyph_group_1:
                    logger.error(
                        f"Multiple kerning groups (kern1) for glyph {g}: "
                        f"{glyph_group_1[g]}, {name}. New group ignored."
                    )
                    continue
                glyph_group_1[g] = name
        elif name.startswith("public.kern2"):
            for g in glyphs:
                if g in glyph_group_2:
                    logger.error(
                        f"Multiple kerning groups (kern2) for glyph {g}: "
                        f"{glyph_group_2[g]}, {name}. New group ignored."
                    )
                    continue
                glyph_group_2[g] = name
    # Rebuild groups to get rid of duplicate kerning groups per glyph and side.
    groups = rebuild_kerning_groups(groups, glyph_group_1, glyph_group_2)
    return groups, glyph_group_1, glyph_group_2


def _build_groups(glyph_to_group_mapping: dict[str, str]) -> "UfoGroups":
    groups: "UfoGroups" = {}
    for name, group in glyph_to_group_mapping.items():
        if group in groups:
            groups[group].append(name)
        else:
            groups[group] = [name]
    return groups


def rebuild_kerning_groups(
    groups: "UfoGroups", glyph_group_1: dict[str, str], glyph_group_2: dict[str, str]
) -> "UfoGroups":
    """
    Rebuild the kerning groups from the left and right side glyph to group mappings.
    This should get rid of any duplicate groups per glyph.
    """
    # Filter groups so only non-kerning groups are kept
    new_groups: "UfoGroups" = {
        n: g for n, g in groups.items() if not n.startswith("public.kern")
    }
    new_groups.update(_build_groups(glyph_group_1))
    new_groups.update(_build_groups(glyph_group_2))
    return new_groups



from math import radians, tan
from typing import TYPE_CHECKING

from vfbLib import DIRECTIONS
from vfbLib.compilers.base import BaseCompiler

if TYPE_CHECKING:
    from vfbLib.typing import GuidePropertiesDict, MMGuidesDict


class GuidesCompiler(BaseCompiler):
    def _compile(self, data: "MMGuidesDict") -> None:
        assert self.vfb is not None
        for direction in DIRECTIONS:
            dir_guides = data[direction]
            self.write_value(len(dir_guides))
            for guide in dir_guides:
                for master_index in range(self.vfb.num_masters):
                    self.write_value(guide[master_index]["pos"])
                    angle = int(tan(radians(guide[master_index]["angle"])) * 10000)
                    self.write_value(angle)


class GuidePropertiesCompiler(BaseCompiler):
    def _compile(self, data: "GuidePropertiesDict") -> None:
        for direction in DIRECTIONS:
            dir_guides = data[direction]
            for gpd in dir_guides:
                self.write_value(gpd["index"])
                if color_rgb := gpd.get("color"):
                    # The hash sign must be stripped, and the components reordered BGR
                    color_bgr = f"{color_rgb[5:]}{color_rgb[3:5]}{color_rgb[1:3]}"
                    value = int(color_bgr, 16)
                    self.write_value(value, shortest=False, signed=False)
                else:
                    self.write_value(-1)
                self.write_str_with_len(gpd.get("name"))
            self.write_value(0)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable

# Length of data types used by StreamReader and Stream Writer etc.
int8_size = 1
int16_size = 2
int32_size = 4
double_size = 8


def binaryToIntList(value: int, start: int = 0) -> list[int]:
    intList = []
    counter = start
    while value:
        if value & 1:
            intList.append(counter)
        value >>= 1
        counter += 1
    return intList


def intListToBinary(bits: "Iterable[int]") -> int:
    value = 0
    for bit in bits:
        value += 2**bit
    return value


def deHexStr(hexstr: str) -> bytes:
    return bytes.fromhex(hexstr)


def hexStr(b: bytes) -> str:
    return bytes.hex(b)



import logging
from argparse import ArgumentParser
from pathlib import Path

from vfbLib.enum import F
from vfbLib.json import save_vfb_json
from vfbLib.ufo.builder import VfbToUfoBuilder
from vfbLib.vfb.vfb import Vfb


def vfb2json():
    parser = ArgumentParser(
        description="vfb2json Converter\nCopyright (c) 2026 by LucasFonts"
    )
    parser.add_argument(
        "-d",
        "--no-decompile",
        action="store_true",
        default=False,
        help="don't decompile data, output binary in JSON",
    )
    parser.add_argument(
        "--header",
        action="store_true",
        default=False,
        help="only read the VFB header, not the actual data",
    )
    parser.add_argument(
        "-m",
        "--minimal",
        action="store_true",
        default=False,
        help="parse only minimal amount of data",
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        nargs=1,
        help="output folder",
    )
    parser.add_argument(
        "-r",
        "--roundtrip",
        action="store_true",
        default=False,
        help="roundtrip data by decompiling and compiling again before saving",
    )
    parser.add_argument(
        "-u",
        "--unicode-strings",
        action="store_true",
        default=False,
        help="force strings to be interpreted as Unicode instead of Windows-1252",
    )
    parser.add_argument(
        "inputpath",
        type=str,
        nargs=1,
        help="input file path (.vfb)",
    )
    args = parser.parse_args()
    if args:
        vfb_path = Path(args.inputpath[0])
        if not vfb_path.exists():
            import sys

            print(f"File not found: '{vfb_path}'", file=sys.stderr)
            sys.exit(1)

        print(parser.description)
        print(f"Reading file {vfb_path} ...")
        save_vfb_json(
            vfb_path,
            out_path=args.path[0] if args.path else None,
            only_header=args.header,
            minimal=args.minimal,
            unicode_strings=args.unicode_strings,
            no_decompile=args.no_decompile,
            roundtrip=args.roundtrip,
        )

    else:
        parser.print_help()


def vfb2ufo():
    parser = ArgumentParser(
        description="vfb3ufo Converter\nCopyright (c) 2026 by LucasFonts"
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        nargs=1,
        help="output folder",
    )
    parser.add_argument(
        "-fo",
        "--force-overwrite",
        action="store_true",
        default=False,
        help="force overwrite",
    )
    parser.add_argument(
        "-g",
        "--keep-groups",
        action="store_true",
        default=False,
        help="don't move non-kerning groups from groups.plist to feature code",
    )
    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        default=False,
        help=(
            "serialize UFO as JSON; "
            "combine with -z to write JSON without line breaks/indentation"
        ),
    )
    parser.add_argument(
        "-k",
        "--add-kerning-groups",
        action="store_true",
        default=False,
        help="add kerning groups to feature code",
    )
    parser.add_argument(
        "-ttx",
        "--ttx",
        action="store_true",
        default=False,
        help="convert binary OpenType Layout data using TTX-like format",
    )
    parser.add_argument(
        "-64",
        "--base64",
        action="store_true",
        default=False,
        help="write GLIF lib 'data' section using base64",
    )
    parser.add_argument(
        "-s",
        "--silent",
        action="store_true",
        default=False,
        help="no display (silent mode)",
    )
    parser.add_argument(
        "-n",
        "--no-normalize",
        action="store_true",
        default=False,
        help="don't normalize the UFOs (faster)",
    )
    parser.add_argument(
        "-nops",
        "--no-postscript-hints",
        action="store_true",
        default=False,
        help="Don't output PostScript hinting",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Print debugging output",
    )
    parser.add_argument(
        "-z",
        "--zip",
        action="store_true",
        default=False,
        help="write UFOZ (compressed UFO)",
    )
    parser.add_argument(
        "inputpath",
        type=str,
        nargs=1,
        help="input file path (.vfb)",
    )
    parser.add_argument(
        "outputpath",
        type=str,
        nargs="?",
        help="output file path (.ufo[z])",
    )
    parser.add_argument(
        "-m",
        "--minimal",
        action="store_true",
        default=False,
        help="parse only minimal amount of data, drop missing glyphs from groups, etc.",
    )
    parser.add_argument(
        "-u",
        "--unicode-strings",
        action="store_true",
        default=False,
        help="force strings to be interpreted as Unicode instead of Windows-1252",
    )
    args = parser.parse_args()
    if args:
        vfb_path = Path(args.inputpath[0])
        if not vfb_path.exists():
            import sys

            print(f"File not found: '{vfb_path}'", file=sys.stderr)
            sys.exit(1)

        if args.verbose:
            logging.basicConfig(level=logging.DEBUG)
        if not args.silent:
            print(parser.description)
            print(f"Reading file {vfb_path} ...")
        vfb = Vfb(
            vfb_path,
            minimal=args.minimal,
            drop_keys={F.Encoding, F.EncodingDefault},
            unicode_strings=args.unicode_strings,
        )
        suffix = ".ufo"
        if args.zip and not args.json:
            suffix += "z"
        if args.path:
            out_path = (Path(args.path[0]) / vfb_path.name).with_suffix(suffix)
        else:
            out_path = vfb_path.with_suffix(suffix)
        vfb.decompile()
        builder = VfbToUfoBuilder(
            vfb,
            minimal=args.minimal,
            base64=args.base64,
            pshints=not args.no_postscript_hints,
            add_kerning_groups=args.add_kerning_groups,
            move_groups=not args.keep_groups,
            normalize=not args.no_normalize,
        )
        builder.write(
            out_path,
            overwrite=args.force_overwrite,
            silent=args.silent,
            ufoz=args.zip,
            json=args.json,
        )
    else:
        parser.print_help()



from struct import pack
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from io import BufferedIOBase


def write_value(value: int, stream: "BufferedIOBase", signed=True) -> None:
    """
    Encode and write an int value to the stream in the shortest possible representation.

    Args:
        value (int): The value to write to the stream.
        stream (BufferedIOBase): The output stream.
        signed (bool, optional): Whether the value is signed. Defaults to True.
    """
    if -107 <= value <= 107:
        # 1-byte representation
        encoded = pack(">B", (value + 0x8B))
    elif 107 < value <= 1131:
        # 2-byte representation
        encoded = pack(">H", (value + 0xF694))
    elif -1131 <= value < -107:
        # 2-byte representation, negative values
        encoded = pack(">H", (-value + 0xFA94))
    else:
        # 5-byte representation
        write_value_long(value, stream, signed)
        return

    stream.write(encoded)


def write_value_long(value: int, stream: "BufferedIOBase", signed=True) -> None:
    """
    Write an integer value to the stream using the longest encoding (4 bytes plus
    marker byte).

    Args:
        value (int): The value to write to the stream.
        stream (BufferedIOBase): The output stream.
        signed (bool, optional): Whether the value is signed. Defaults to True.
    """
    fmt = "i" if signed else "I"
    encoded = pack(f">B{fmt}", 0xFF, value)
    stream.write(encoded)



import logging
from struct import unpack
from typing import TYPE_CHECKING

from vfbLib import tt_settings, ttinfo_names
from vfbLib.helpers import binaryToIntList
from vfbLib.parsers.base import BaseParser
from vfbLib.typing import (
    TrueTypeInfoDict,
    TTStemDict,
    TTStemsDict,
    TTZoneDict,
    TTZonesDict,
    VdmxRecDict,
)

if TYPE_CHECKING:
    from vfbLib.typing import FlagsOptionsDict, GaspList

logger = logging.getLogger(__name__)


def convert_int_to_flags_options(value: int) -> "FlagsOptionsDict":
    flags = binaryToIntList(value & 0xFFFF)
    options = binaryToIntList(value >> 16)
    return {
        "flags": flags,
        "options": [tt_settings.get(i, str(i)) for i in options],
    }


class GaspParser(BaseParser):
    """
    A parser that reads data as an array representing Gasp table values.
    """

    def _parse(self) -> "GaspList":
        data = self.stream.read()
        gasp = unpack(f"<{len(data) // 2}H", data)
        it = iter(gasp)
        return [
            {
                "maxPpem": a,
                "flags": b,
            }
            for a, b in zip(it, it)
        ]


class TrueTypeInfoParser(BaseParser):
    """
    A parser that reads data as "TrueType Info" values.
    """

    def assert_unique(self, d: TrueTypeInfoDict, key: str) -> None:
        if key in d:
            logger.error(f"Duplicate key in TrueType Info: {key} in {d}")
            raise KeyError

    def _parse(self) -> TrueTypeInfoDict:
        info: TrueTypeInfoDict = TrueTypeInfoDict()

        while True:
            k = self.read_uint8()
            bk = ttinfo_names.get(k, f"bit{k}")  # dict key, as a bit number
            dk = ttinfo_names.get(k, str(k))  # dict key, human-readable
            hk = ttinfo_names.get(k, hex(k))  # dict key, hex

            match k:
                case 0x32:
                    # End
                    return info

                case 0x33 | 0x34 | 0x35 | 0x36 | 0x37 | 0x38:
                    self.assert_unique(info, dk)
                    info[dk] = self.read_value()

                case 0x39:
                    # Options and head flags
                    self.assert_unique(info, dk)
                    info[dk] = convert_int_to_flags_options(self.read_value())

                case 0x3A | 0x3B | 0x3C | 0x3D | 0x3E | 0x3F:
                    # head: upm, mac_style, lowest_rec_ppm, font_direction_hint
                    # OS/2: weight_class, width_class
                    self.assert_unique(info, bk)
                    info[bk] = self.read_value()

                case (
                    0x40
                    | 0x41
                    | 0x42
                    | 0x43
                    | 0x44
                    | 0x45
                    | 0x46
                    | 0x47
                    | 0x48
                    | 0x49
                    | 0x4A
                    | 0x4B
                ):
                    # OS/2: fs_type, sub/superscripts, strikeout, family_class
                    self.assert_unique(info, dk)
                    info[dk] = self.read_value()

                case 0x4C:
                    # PANOSE
                    self.assert_unique(info, dk)
                    info[dk] = [self.read_uint8() for _ in range(10)]

                case 0x4D | 0x4E | 0x4F | 0x50 | 0x51 | 0x52:
                    # OS/2: typo_ascender, typo_descender, typo_line_gap, fs_selection,
                    # win_ascent, win_descent
                    self.assert_unique(info, dk)
                    info[dk] = self.read_value()

                case 0x53:
                    # HDMX ppms 1
                    self.assert_unique(info, dk)
                    num_values = self.read_value()
                    info[dk] = [self.read_uint8() for _ in range(num_values)]

                case 0x54:
                    # Codepages
                    self.assert_unique(info, dk)
                    range1 = self.read_value(signed=False)
                    range2 = self.read_value(signed=False)
                    info[dk] = {
                        "os2_ul_code_page_range1": range1,
                        "os2_ul_code_page_range2": range2,
                    }

                case 0x56 | 0x57:
                    # 0x56: head_creation
                    # 0x57: value 2 of the list that FL5 returns for
                    #           ttinfo.head_creation
                    # The timestamps are returned by the FL5 API as signed, so for
                    # current dates they have wrapped to a negative number.
                    self.assert_unique(info, dk)
                    info[dk] = self.read_value(signed=True)

                case 0x5C:
                    # Average width
                    self.assert_unique(info, dk)
                    info[dk] = self.read_value()

                case 0x58:
                    # HDMX ppms 2
                    self.assert_unique(info, hk)
                    num_values = self.read_value()
                    info[hk] = [self.read_uint8() for _ in range(num_values)]

                case _:
                    logger.warning(f"Unknown key in TrueType info: {hex(k)}")


class TrueTypeStemsParser(BaseParser):
    def _parse(self):
        names = ("ttStemsV", "ttStemsH")
        result = TTStemsDict(ttStemsV=[], ttStemsH=[])
        for i in range(2):
            direction = []
            num_stems = self.read_value()
            for _ in range(num_stems):
                width = self.read_value()
                stem_name_length = self.read_uint8()
                stem_name = self.stream.read(stem_name_length).decode(self.encoding)
                ppm6 = self.read_value()

                direction.append(
                    TTStemDict(value=width, name=stem_name, round={6: ppm6})
                )

            result[names[i]] = direction

        return result


class TrueTypeStemPpemsParser(BaseParser):
    def _parse(self) -> TTStemsDict:
        names = ("ttStemsV", "ttStemsH")
        result = TTStemsDict(ttStemsV=[], ttStemsH=[])
        for i in range(2):
            direction = []
            num_stems = self.read_value()
            d = {}
            for j in range(num_stems):
                for k in range(2, 6):
                    ppm = self.read_value()
                    d[k] = ppm

                direction.append(TTStemDict(stem=j, round=d.copy()))

            result[names[i]] = direction

        return result


class TrueTypeStemPpems1Parser(BaseParser):
    # PPEM 1 for each stem is stored in a separate entry ...
    def _parse(self) -> TTStemsDict:
        assert self.vfb is not None
        names = ("ttStemsV", "ttStemsH")
        result = TTStemsDict(ttStemsV=[], ttStemsH=[])
        for i in range(2):
            direction = []
            num_stems = (self.vfb.ttStemsV_count, self.vfb.ttStemsH_count)[i]
            if num_stems is None:
                raise ValueError

            for j in range(num_stems):
                ppm = self.read_value()
                direction.append(TTStemDict(stem=j, round={1: ppm}))

            result[names[i]] = direction

        return result


class TrueTypeStemPpems23Parser(BaseParser):
    # PPEM 2 and 3 for each stem are stored in a separate entry ... sometimes!?
    def _parse(self) -> TTStemsDict:
        names = ("ttStemsV", "ttStemsH")
        result = TTStemsDict(ttStemsV=[], ttStemsH=[])
        for i in range(2):
            direction = []
            num_stems = self.read_value()
            if num_stems is None:
                raise ValueError

            for j in range(num_stems):
                ppm2 = self.read_value()
                ppm3 = self.read_value()
                direction.append(TTStemDict(stem=j, round={2: ppm2, 3: ppm3}))

            result[names[i]] = direction

        return result


class TrueTypeZoneDeltasParser(BaseParser):
    def _parse(self) -> dict[int, dict[int, int]]:
        num_deltas = self.read_value()
        result: dict[int, dict[int, int]] = {}
        for _ in range(num_deltas):
            # Index into Bottom + Top Zones
            index = self.read_value()
            ppm = self.read_value()
            shift = self.read_value()
            if index in result:
                result[index][ppm] = shift
            else:
                result[index] = {ppm: shift}

        return result


class TrueTypeZonesParser(BaseParser):
    def _parse(self) -> TTZonesDict:
        names = ("ttZonesT", "ttZonesB")
        result = TTZonesDict(ttZonesT=[], ttZonesB=[])
        for i in range(2):
            side = []
            num_zones = self.read_value()
            logger.debug(f"Zones: {num_zones}")
            for _ in range(num_zones):
                position = self.read_value()
                width = self.read_value()
                zone_name = self.read_str_with_len()
                side.append(TTZoneDict(position=position, value=width, name=zone_name))

            result[names[i]] = side

        return result


class VdmxParser(BaseParser):
    def _parse(self) -> list[VdmxRecDict]:
        result = []
        num_records = self.read_value()
        for _ in range(num_records):
            rec = VdmxRecDict(
                pelHeight=self.read_value(),
                max=self.read_value(),
                min=self.read_value(),
            )
            result.append(rec)

        return result



from vfbLib.parsers.base import BaseParser
from vfbLib.typing import PCLTDict


class PcltParser(BaseParser):
    """
    A parser that reads the PCLT table data
    """

    def _parse(self) -> PCLTDict:
        values: PCLTDict = PCLTDict()

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#fontnumber
        values["font_number"] = self.read_value(signed=False)

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#pitch
        values["pitch"] = self.read_value()

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#xheight
        values["x_height"] = self.read_value()

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#style
        values["style"] = self.read_value()

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#typefamily
        values["type_family"] = self.read_value()

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#capheight
        values["cap_height"] = self.read_value()

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#symbolset
        values["symbol_set"] = self.read_value()

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#typeface
        values["typeface"] = self.read_str(16)

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#charactercomplement
        values["character_complement"] = [self.read_uint8() for _ in range(8)]

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#filename
        values["file_name"] = self.read_str(6)

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#strokeweight
        values["stroke_weight"] = self.read_int8()

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#widthtype
        values["width_type"] = self.read_int8()

        # https://learn.microsoft.com/de-de/typography/opentype/spec/pclt#serifstyle
        values["serif_style"] = self.read_uint8()

        return values



from typing import TYPE_CHECKING

from vfbLib.parsers.base import StreamReader

if TYPE_CHECKING:
    from io import BufferedIOBase

    from vfbLib.typing import VfbHeaderDict

# FontLab 3.0 and newer
FL30_SIGNATURE = 0x464C571A  # _WLF

FL30_APP = 0x31
FL30_FILE = 0x30

# FontLab 2.5 (not implemented here)
FL25_SIGNATURE = 0x53564EFF
FL25_APP = 10
FL25_FILE = 13


class VfbHeaderParser(StreamReader):
    def __init__(self, stream: "BufferedIOBase") -> None:
        self.stream = stream

    def parse(self) -> "VfbHeaderDict":
        signature = self.read_uint32()
        assert signature == FL30_SIGNATURE, f"File signature not supported: {signature}"
        app_version = self.read_uint8()
        assert app_version == FL30_APP, f"Unsupported app version: {app_version}"
        file_version = self.read_uint8()
        assert file_version == FL30_FILE, f"Unsupported file version: {file_version}"
        version_major = self.read_uint8()
        assert version_major == 3, f"Unsupported app version (major): {version_major}"
        version_minor = self.read_uint8()

        header: "VfbHeaderDict" = {
            "signature": signature,
            "app_version": app_version,
            "file_version": file_version,
            "version_major": version_major,
            "version_minor": version_minor,
        }

        data_offset = self.read_uint32()
        self.stream.seek(data_offset)

        return header



from typing import Any

from vfbLib.compilers.base import BaseCompiler
from vfbLib.helpers import deHexStr


class BinaryTableCompiler(BaseCompiler):
    """
    A compiler that compiles binary table data.
    """

    def _compile(self, data: Any) -> None:
        self.write_str(data["tag"])  # FIXME: Add padding here?
        self.stream.write(deHexStr(data["data"]))



from typing import Any, Literal, NotRequired, TypedDict

Point = tuple[int, int]


class AnchorDict(TypedDict):
    name: NotRequired[str]
    x: int
    y: int


class AnchorPropertiesDict(TypedDict):
    hue: NotRequired[int]
    data: NotRequired[int]


KerningClassFlagDict = dict[str, tuple[int, int]]

MetricsClassFlagDict = dict[str, tuple[int, int, int]]


class BitmapDataDict(TypedDict):
    data: list[list[int]]
    preview: NotRequired[list[str]]


class BackgroundImageDict(TypedDict):
    origin: tuple[int, int]
    size_units: tuple[int, int]
    size_pixels: tuple[int, int]
    bitmap: BitmapDataDict


class BBoxDict(TypedDict):
    xMin: int
    yMin: int
    xMax: int
    yMax: int


class Component(TypedDict):
    gid: int
    offsetX: list[int]
    offsetY: list[int]
    scaleX: list[float]
    scaleY: list[float]


class CustomCmap(TypedDict):
    language_id: int
    platform_id: int
    encoding_id: int
    format: int
    option: int
    page_name: str


EntryDecompiled = dict[str, Any] | int | list[Any] | str | tuple[int, str] | None


class EntryDict(TypedDict):
    key: str
    size: NotRequired[int]
    data: bytes | EntryDecompiled
    parser: NotRequired[str]
    compiler: NotRequired[str]


class ExpandKernFlagsDict(TypedDict):
    limit_action: int
    limit_codepage: int
    limit_cmap_10: int
    limit_font_window: int
    limit_count: int
    limit_keep: int
    apply_to_assistance: int


class FeatureDict(TypedDict):
    tag: str
    code: list[str]


class FeaturesDict(TypedDict):
    prefix: list[str]
    features: list[FeatureDict]


class FlagsOptionsDict(TypedDict):
    flags: list[int]
    options: list[str]


class FLVersionDict(TypedDict):
    platform: str
    version: tuple[int, ...]
    owner: int


class TTAutoHintOptionsDict(TypedDict):
    single_link_attachment_precision: int
    generate_triple_hints: int
    generate_delta_instructions: int
    direct_links_to_center_of_the_glyph_where_possible: int
    interpolate_positions_of_cusp_points: int
    interpolate_positions_of_double_links: int
    add_link_to_rsb: int


class FontOptionsDict(TypedDict):
    fit_ascender: NotRequired[int]
    fit_descender: NotRequired[int]
    auto_metrics_left: NotRequired[int]
    auto_metrics_right: NotRequired[int]
    auto_metrics_close: NotRequired[int]
    auto_hinting_min_h_len: NotRequired[int]
    auto_hinting_min_v_len: NotRequired[int]
    auto_hinting_min_h_width: NotRequired[int]
    auto_hinting_min_v_width: NotRequired[int]
    auto_hinting_max_h_width: NotRequired[int]
    auto_hinting_max_v_width: NotRequired[int]
    auto_hinting_h_ratio: NotRequired[float]
    auto_hinting_v_ratio: NotRequired[float]
    duplicate_place_x: NotRequired[int]
    paste_place_x: NotRequired[int]
    duplicate_place_y: NotRequired[int]
    paste_place_y: NotRequired[int]
    opentype_name_records: NotRequired[int]
    codepage_for_cmap_1_0: NotRequired[int]
    dont_ignore_unicode_indexes: NotRequired[int]
    head_bbox_savings: NotRequired[int]
    autohinting_options: NotRequired[int | TTAutoHintOptionsDict]
    export_hinted_truetype_font: NotRequired[int]
    autohint_unhinted_glyphs: NotRequired[int]
    keep_existing_truetype_instructions: NotRequired[int]
    export_visual_truetype_hints: NotRequired[int]
    apply_bbox_savings: NotRequired[int]
    auto_win_asc_desc: NotRequired[int]
    add_characters: NotRequired[int]
    export_embedded_bitmaps: NotRequired[int]
    copy_hdmx_data_from_base_to_composite_glyph: NotRequired[int]
    dont_automatically_reorder_glyphs: NotRequired[int]
    export_ot: NotRequired[int]
    export_volt: NotRequired[int]
    write_kern_feature: NotRequired[int]
    t1_terminal: NotRequired[int]
    t1_pfm: NotRequired[int]
    t1_afm: NotRequired[int]
    t1_autohint: NotRequired[int]
    t1_unicode: NotRequired[int]
    optimize_align: NotRequired[int]
    optimize_reduce: NotRequired[int]
    t1_encoding: NotRequired[int]
    ot_write_gdef: NotRequired[int]
    t1_use_os2: NotRequired[int]
    subrize: NotRequired[int]
    t1_sort: NotRequired[int]
    export_kern_table: NotRequired[int]
    t1_fs_type: NotRequired[int]
    expand_kern_flags: NotRequired[int | ExpandKernFlagsDict]
    expand_kern_codepage: NotRequired[int]
    expand_kern_count: NotRequired[int]
    decompose: NotRequired[int]


class GaspRangeDict(TypedDict):
    maxPpem: int
    flags: int


GaspList = list[GaspRangeDict]


class GdefDict(TypedDict):
    anchors: NotRequired[list[AnchorDict]]
    carets: NotRequired[list[tuple[int, int]]]
    glyph_class: NotRequired[str]
    ot_classes: NotRequired[list[int]]


class GlyphBitmapDict(TypedDict):
    ppm: int
    origin: tuple[int, int]
    adv: tuple[int, int]
    size_pixels: tuple[int, int]
    bitmap: BitmapDataDict


class GlyphHintingOptionsDict(TypedDict):
    hint_replacement: NotRequired[int]
    horizontal_3_stem: NotRequired[int]
    vertical_3_stem: NotRequired[int]
    other: NotRequired[list[int]]


class GuideDict(TypedDict):
    angle: float | int
    pos: int


GuideList = list[GuideDict]


class MMGuidesDict(TypedDict):
    h: list[GuideList]
    v: list[GuideList]


class GuidePropertyDict(TypedDict):
    color: NotRequired[str]
    index: int
    name: NotRequired[str]


class GuidePropertiesDict(TypedDict):
    h: list[GuidePropertyDict]
    v: list[GuidePropertyDict]


class HintDict(TypedDict):
    pos: int
    width: int


HintTuple = tuple[str, int, int]


class Instruction(TypedDict):
    cmd: str
    params: dict[str, int]


class LinkDict(TypedDict):
    x: list[tuple[int, int]]
    y: list[tuple[int, int]]


class MappingModeDict(TypedDict):
    mapping_mode: str
    m2: int
    m3: int
    mapping_id: int


class MMAnchorDict(TypedDict):
    x: list[int]
    y: list[int]


class MMHintsDict(TypedDict):
    h: list[list[HintDict]]
    v: list[list[HintDict]]
    hintmasks: NotRequired[list[tuple[str, int]]]


class MMNode(TypedDict):
    flags: int
    points: list[list[Point]]
    type: Literal["move", "line", "curve", "qcurve"]


class PCLTDict(TypedDict):
    font_number: int
    pitch: int
    x_height: int
    style: int
    type_family: int
    cap_height: int
    symbol_set: int
    typeface: str
    character_complement: list[int]
    file_name: str
    stroke_weight: int
    width_type: int
    serif_style: int


class PrimaryInstanceDict(TypedDict):
    name: str
    values: tuple[float, ...]


class PSInfoDict(TypedDict):
    font_matrix: tuple[float, ...]
    force_bold: int
    blue_values: list[int]  # max 14
    other_blues: list[int]  # max 10
    family_blues: list[int]  # max 14
    family_other_blues: list[int]  # max 10
    blue_scale: float
    blue_shift: int
    blue_fuzz: int
    std_hw: int
    std_vw: int
    stem_snap_h: list[int]
    stem_snap_v: list[int]
    bounding_box: BBoxDict
    adv_width_min: int
    adv_width_max: int
    adv_width_avg: int
    ascender: int
    descender: int
    x_height: int
    cap_height: int


class CodepagesDict(TypedDict):
    os2_ul_code_page_range1: int
    os2_ul_code_page_range2: int


class TrueTypeInfoDict(TypedDict):
    max_zones: int
    max_twilight_points: int
    max_storage: int
    max_function_defs: int
    max_instruction_defs: int
    max_stack_elements: int
    head_flags: int | FlagsOptionsDict
    head_units_per_em: int
    head_mac_style: int
    head_lowest_rec_ppem: int
    head_creation: int
    head_creation2: int
    head_font_direction_hint: int
    os2_us_weight_class: int
    os2_us_width_class: int
    os2_fs_type: int
    os2_y_subscript_x_size: int
    os2_y_subscript_y_size: int
    os2_y_subscript_x_offset: int
    os2_y_subscript_y_offset: int
    os2_y_superscript_x_size: int
    os2_y_superscript_y_size: int
    os2_y_superscript_x_offset: int
    os2_y_superscript_y_offset: int
    os2_y_strikeout_size: int
    os2_y_strikeout_position: int
    os2_s_family_class: int
    OpenTypeOS2Panose: list[int]
    os2_s_typo_ascender: int
    os2_s_typo_descender: int
    os2_s_typo_line_gap: int
    os2_fs_selection: int
    os2_us_win_ascent: int
    os2_us_win_descent: int
    AverageWidth: int
    HdmxPPMs1: list[int]
    HdmxPPMs2: list[int]
    Codepages: CodepagesDict
    hhea_line_gap: NotRequired[int]
    hhea_ascender: NotRequired[int]
    hhea_descender: NotRequired[int]


class TTCommandDict(TypedDict):
    name: str
    params: list[str]


class TTStemDict(TypedDict):
    name: NotRequired[str]
    round: dict[int, int]
    stem: NotRequired[int]
    value: NotRequired[int]


class TTStemsDict(TypedDict):
    ttStemsV: list[TTStemDict]
    ttStemsH: list[TTStemDict]


class TTZoneDict(TypedDict):
    deltas: NotRequired[dict[int, int]]
    name: str
    position: int
    top: NotRequired[bool]
    value: int


class TTZonesDict(TypedDict):
    ttZonesT: list[TTZoneDict]
    ttZonesB: list[TTZoneDict]


class GlyphData(TypedDict):
    components: NotRequired[list[Component]]
    # constants: NotRequired[Tuple[Any, ...]]
    guides: NotRequired[MMGuidesDict]
    hints: NotRequired[MMHintsDict]
    imported: NotRequired[Any]  # FIXME
    kerning: NotRequired[dict[int, list[int]]]
    metrics: NotRequired[list[Point]]
    name: NotRequired[str]
    nodes: NotRequired[list[MMNode]]
    num_masters: NotRequired[int]
    # num_node_values: NotRequired[int]
    tth: NotRequired[list[Instruction]]


class MaskData(GlyphData):
    weight_vector: list[float]


class VdmxRecDict(TypedDict):
    pelHeight: int
    max: int
    min: int


class VfbDict(TypedDict):
    header: "VfbHeaderDict"
    entries: list[dict[str, Any]]


class VfbHeaderDict(TypedDict):
    signature: int
    app_version: int
    file_version: int
    version_major: int
    version_minor: int



import logging
from functools import cached_property
from typing import TYPE_CHECKING, Any

from ufoLib2.objects.info import Info, WidthClass

from vfbLib.enum import F, T
from vfbLib.helpers import binaryToIntList
from vfbLib.ufo.time import convert_timestamp
from vfbLib.ufo.typing import TUfoGaspRecDict
from vfbLib.ufo.vfb2ufo import TT_UFO_LIB_KEY

if TYPE_CHECKING:
    from vfbLib.typing import GaspList
    from vfbLib.ufo.typing import UfoGuide


logger = logging.getLogger(__name__)


maxp_keys = {
    "max_zones": "maxZones",
    "max_twilight_points": "maxTwilightPoints",
    "max_storage": "maxStorage",
    "max_function_defs": "maxFunctionDefs",
    "max_instruction_defs": "maxInstructionDefs",
    "max_stack_elements": "maxStackElements",
}

default_tt_lib = {
    "formatVersion": "1",
    "maxZones": 0,
    "maxTwilightPoints": 0,
    "maxStorage": 0,
    "maxFunctionDefs": 0,
    "maxInstructionDefs": 0,
    "maxStackElements": 0,
}


class VfbToUfoInfo(Info):
    def __init__(self) -> None:
        super().__init__()
        # Chance to set some defaults that should always be written
        self.familyName = "Untitled"
        self.guidelines: "list[UfoGuide]" = []
        self.italicAngle: float | int = 0
        self.openTypeGaspRangeRecords: list[TUfoGaspRecDict] = []
        self.openTypeHeadCreated: str | None = None
        self.openTypeHeadFlags: list[int] = []
        self.openTypeHheaLineGap = 0
        self.openTypeHeadLowestRecPPEM = 0
        self.openTypeNameDescription = ""
        self.openTypeNamePreferredFamilyName = ""
        self.openTypeNamePreferredSubfamilyName = ""
        self.openTypeNameRecords: list | None = None
        self.openTypeNameSampleText = ""
        self.openTypeOS2FamilyClass = [0, 0]
        self.openTypeOS2Selection: list[int] = []
        self.openTypeOS2SubscriptXSize = 0
        self.openTypeOS2SubscriptYSize = 0
        self.openTypeOS2SubscriptXOffset = 0
        self.openTypeOS2SubscriptYOffset = 0
        self.openTypeOS2SuperscriptXSize = 0
        self.openTypeOS2SuperscriptYSize = 0
        self.openTypeOS2SuperscriptXOffset = 0
        self.openTypeOS2SuperscriptYOffset = 0
        self.openTypeOS2StrikeoutPosition = 0
        self.openTypeOS2StrikeoutSize = 0
        self.openTypeOS2Type: list[int] = []
        self.openTypeOS2UnicodeRanges: list[int] | None = []
        self.openTypeOS2WeightClass = 400
        self.openTypeOS2WidthClass = WidthClass(5)
        self.postscriptBlueValues: list[float] | None = []
        self.postscriptDefaultWidthX: float | None = None
        self.postscriptFamilyBlues: list[float] | None = []
        self.postscriptFamilyOtherBlues: list[float] | None = []
        self.postscriptFontName = "Untitled-Regular"
        self.postscriptForceBold = False
        self.postscriptFullName = "Untitled Regular"
        self.postscriptIsFixedPitch = False
        self.postscriptNominalWidthX: float | None = None
        self.postscriptOtherBlues: list[float] | None = []
        self.postscriptStemSnapH: list[float] | None = None
        self.postscriptStemSnapV: list[float] | None = None
        self.postscriptUnderlinePosition = 0
        self.postscriptUnderlineThickness = 0
        self.postscriptWeightName = "Regular"
        self.styleMapFamilyName = "Untitled"
        self.styleMapStyleName = "regular"
        self.styleName = "Regular"

        # Add sensible defaults?
        self.openTypeOS2StrikeoutSize = 0
        self.openTypeOS2StrikeoutPosition = 0
        self.openTypeOS2TypoAscender = 0
        self.openTypeOS2TypoDescender = 0
        self.openTypeOS2TypoLineGap = 0
        self.openTypeOS2WinAscent = 0
        self.openTypeOS2WinDescent = 0
        self.openTypeOS2CodePageRanges: list[int] = []

        # Some lib values are gathered here, they are added to the UFO lib in builder
        self.lib = {}

        self.build_mapping()

    @cached_property
    def ds_family_name(self) -> str | None:
        """
        Return the family name for use in the DesignSpaceDocument.
        """
        if self.openTypeNamePreferredFamilyName:
            return self.openTypeNamePreferredFamilyName
        if self.familyName:
            return self.familyName
        return self.styleMapFamilyName

    @cached_property
    def ds_style_name(self) -> str | None:
        """
        Return the style name for use in the DesignSpaceDocument.
        """
        if self.openTypeNamePreferredSubfamilyName:
            return self.openTypeNamePreferredSubfamilyName
        if self.styleName:
            return self.styleName
        return self.styleMapStyleName

    @cached_property
    def ui_name(self) -> str:
        name = ""
        if hasattr(self, "familyName"):
            name = self.familyName

        elif hasattr(self, "postscriptFontName"):
            name = self.postscriptFontName
        if name:
            return name
        return "Untitled Regular"

    def build_mapping(self):
        self.mapping = {
            # F.pref_family_name: "openTypeNamePreferredFamilyName",  # below
            F.full_name: "postscriptFullName",
            F.font_name: "postscriptFontName",
            F.family_name: "styleMapFamilyName",
            F.weight: "weightName",
            F.italic_angle: "italicAngle",
            F.underline_position: "postscriptUnderlinePosition",
            F.underline_thickness: "postscriptUnderlineThickness",
            # F.is_fixed_pitch: "postscriptIsFixedPitch",  # below
            F.copyright: "copyright",
            F.notice: "openTypeNameDescription",
            F.source: "openTypeNameManufacturer",
            F.unique_id: "postscriptUniqueID",
            # F.weight_code, below
            F.trademark: "trademark",
            F.designer: "openTypeNameDesigner",
            F.designer_url: "openTypeNameDesignerURL",
            F.vendor_url: "openTypeNameManufacturerURL",
            F.width: "widthName",
            F.License: "openTypeNameLicense",
            F.LicenseURL: "openTypeNameLicenseURL",
            F.fond_id: "macintoshFONDFamilyID",
            F.apple_name: "macintoshFONDName",
            F.panose: "openTypeOS2Panose",
            F.SampleText: "openTypeNameSampleText",
            F.vendor: "openTypeOS2VendorID",
            F.tt_u_id: "openTypeNameUniqueID",
            F.tt_version: "openTypeNameVersion",
            F.version_major: "versionMajor",
            F.version_minor: "versionMinor",
            F.year: "year",
            F.upm: "unitsPerEm",
            # F.pref_style_name: "openTypeNamePreferredSubfamilyName",  # below
            T.hhea_ascender: "openTypeHheaAscender",
            T.hhea_descender: "openTypeHheaDescender",
            T.hhea_line_gap: "openTypeHheaLineGap",
            F.note: "note",
            F.default_character: "postscriptDefaultCharacter",
        }
        # Integer values from TTInfo
        self.mapping_int = {
            # "max_zones": ,
            # "max_twilight_points": "",
            # "max_storage": "",
            # "max_function_defs": "",
            # "max_instruction_defs": "",
            # "max_stack_elements": "",
            # "head_flags": "", # below
            # "head_units_per_em": "unitsPerEm",  # duplicate
            # "os2_us_weight_class": "openTypeOS2WeightClass",  # duplicate
            "os2_us_width_class": "openTypeOS2WidthClass",
            "head_lowest_rec_ppem": "openTypeHeadLowestRecPPEM",
            "os2_y_subscript_x_size": "openTypeOS2SubscriptXSize",
            "os2_y_subscript_y_size": "openTypeOS2SubscriptYSize",
            "os2_y_subscript_x_offset": "openTypeOS2SubscriptXOffset",
            "os2_y_subscript_y_offset": "openTypeOS2SubscriptYOffset",
            "os2_y_superscript_x_size": "openTypeOS2SuperscriptXSize",
            "os2_y_superscript_y_size": "openTypeOS2SuperscriptYSize",
            "os2_y_superscript_x_offset": "openTypeOS2SuperscriptXOffset",
            "os2_y_superscript_y_offset": "openTypeOS2SuperscriptYOffset",
            "os2_y_strikeout_size": "openTypeOS2StrikeoutSize",
            "os2_y_strikeout_position": "openTypeOS2StrikeoutPosition",
            "os2_s_typo_ascender": "openTypeOS2TypoAscender",
            "os2_s_typo_descender": "openTypeOS2TypoDescender",
            "os2_s_typo_line_gap": "openTypeOS2TypoLineGap",
            "os2_us_win_ascent": "openTypeOS2WinAscent",
            "os2_us_win_descent": "openTypeOS2WinDescent",
            # "Average Width": "postscriptNominalWidthX",
        }

    def fix_underline_position(self):
        # VFB stores middle of line and thickness, but spec says it must be
        # stored as top of line and thickness.
        if (
            self.postscriptUnderlinePosition is None
            or self.postscriptUnderlineThickness is None
        ):
            logger.error(
                "Can't fix underline position because position or thickness is None"
            )
            return

        self.postscriptUnderlinePosition += int(
            round(0.5 * self.postscriptUnderlineThickness)
        )

    def set_attr(self, attr: str, data):
        setattr(self, attr, data)

    def set_created_timestamp(self, data: int) -> None:
        self.openTypeHeadCreated = convert_timestamp(data).strftime("%Y/%m/%d %H:%M:%S")

    def set_fixed_pitch(self, data: int) -> None:
        self.postscriptIsFixedPitch = bool(data)

    def set_name_records(self, data: list[tuple[int, int, int, int, str]]) -> None:
        # We need to set the attribute once at the end to be able to write JSON, cf.
        # https://github.com/fonttools/ufoLib2/issues/203
        name_records = []
        for rec in data:
            nameID, platformID, encodingID, languageID, s = rec
            name_records.append(
                {
                    "nameID": nameID,
                    "platformID": platformID,
                    "encodingID": encodingID,
                    "languageID": languageID,
                    "string": s,
                }
            )
        self.openTypeNameRecords = name_records

    def set_selection(self, data: int) -> None:
        # Bit 0 = Italic
        # Bit 5 = Bold
        # Bit 6 = Regular

        # Those should not be included in the list.
        # Any others could remain, but FL doesn't set them.
        intlist = binaryToIntList(data)
        self.openTypeOS2Selection = [b for b in intlist if b not in (0, 5, 6)]

        # Construct the style map style name

        name_parts = []
        if 6 in intlist:
            name_parts.append("regular")
        elif 5 in intlist:
            name_parts.append("bold")
        if 0 in intlist:
            name_parts.append("italic")
        if name_parts:
            self.styleMapStyleName = " ".join(name_parts)

    def set_style_name(self, data: str) -> None:
        self.styleName = data
        self.openTypeNamePreferredSubfamilyName = data

    def set_tt_gasp(self, data: "GaspList") -> None:
        # FIXME: Different results when hinting is compiled with flufocompile
        gasp: list[TUfoGaspRecDict] = []
        for rec in data:
            gasp.append(
                TUfoGaspRecDict(
                    rangeMaxPPEM=rec["maxPpem"],
                    rangeGaspBehavior=binaryToIntList(rec["flags"]),
                )
            )
        self.openTypeGaspRangeRecords = gasp

    def set_tt_info(self, data: dict[str, int | list[int] | list[str]]) -> None:  # noqa: C901, E501
        for k, v in data.items():
            if isinstance(v, int):
                if k in self.mapping_int:
                    setattr(self, self.mapping_int[k], v)
                elif k == "head_creation":
                    self.set_created_timestamp(v)
                elif k == "head_font_direction_hint":
                    # self.info.openTypeOS2Type = binaryToIntList(v)
                    pass
                elif k == "os2_fs_type":
                    self.openTypeOS2Type = binaryToIntList(v)
                elif k == "os2_s_family_class":
                    c = v >> 8
                    s = v & ~(c << 8)
                    self.openTypeOS2FamilyClass = [c, s]
                elif k == "os2_us_weight_class":
                    # Duplicate, we set it from a separate entry (weight_code)
                    # self.openTypeOS2WeightClass = v
                    pass
                elif k == "os2_fs_selection":
                    # Duplicate, set from a separate entry (font_style)
                    # self.openTypeOS2Selection = binaryToIntList(v)
                    pass
                elif k in maxp_keys:
                    # Contains useful values in imported binary fonts
                    self.set_tt_instructions(maxp_keys[k], v)
                elif k == "Average Width":
                    if v != 0:
                        self.postscriptNominalWidthX = abs(v)
                else:
                    logger.info(f"Unhandled int value in UFO info: {k, v}")
            elif isinstance(v, list):
                if k == "OpenTypeOS2Panose":
                    # Duplicate?
                    # if v != self.info.openTypeOS2Panose:
                    #     print("Contradictory PANOSE values")
                    #     print(self.info.openTypeOS2Panose, "vs.", v)
                    pass
                else:
                    logger.info(f"Unhandled list value in UFO info: {k, v}")
            elif isinstance(v, dict):
                if k == "head_flags":
                    self.openTypeHeadFlags = v.get("flags", [])
                elif k == "Codepages":
                    ranges = binaryToIntList(v.get("os2_ul_code_page_range1", 0))
                    for cp in binaryToIntList(v.get("os2_ul_code_page_range2", 0)):
                        ranges.append(cp + 32)
                    if ranges:
                        self.openTypeOS2CodePageRanges = ranges
                else:
                    logger.info(f"Unhandled dict value in UFO info: {k, v}")
            else:
                raise TypeError

    def set_weight_class(self, data: int) -> None:
        self.openTypeOS2WeightClass = min(max(1, data), 1000)

    def set_tt_instructions(self, key: str, value: Any) -> None:
        if TT_UFO_LIB_KEY not in self.lib:
            # Prepare UFO-level lib entry for TrueType values
            self.lib[TT_UFO_LIB_KEY] = {"formatVersion": "1"}
        self.lib[TT_UFO_LIB_KEY][key] = value



from __future__ import annotations

from typing import TYPE_CHECKING

from vfbLib.compilers.base import BaseCompiler

if TYPE_CHECKING:
    from vfbLib.typing import PrimaryInstanceDict


class AnisotropicInterpolationsCompiler(BaseCompiler):
    def _compile(self, data: list[list[tuple[int, int]]]) -> None:
        for axis_mappings in data:
            self.write_value(len(axis_mappings))
            for src, tgt in axis_mappings:
                self.write_value(src)
                self.write_value(tgt)


class AxisMappingsCountCompiler(BaseCompiler):
    def _compile(self, data: list[int]) -> None:
        for value in data:
            self.write_uint32(value)


class AxisMappingsCompiler(BaseCompiler):
    def _compile(self, data: list[tuple[float, float]]) -> None:
        for src, tgt in data:
            self.write_double(src)
            self.write_double(tgt)


class MasterLocationCompiler(BaseCompiler):
    def _compile(self, data: tuple[int, tuple[float, float, float, float]]) -> None:
        master_index, location = data
        self.write_uint32(master_index)
        self.write_doubles(location)


class PrimaryInstancesCompiler(BaseCompiler):
    def _compile(self, data: "list[PrimaryInstanceDict]") -> None:
        self.write_value(len(data))
        for instance in data:
            name = instance["name"]
            self.write_str_with_len(name)
            for value in instance["values"]:
                self.write_value(int(10000 * value))



from re import search

from vfbLib.parsers.base import BaseParser
from vfbLib.typing import FeaturesDict


class NameRecordsParser(BaseParser):
    def _parse(self) -> list[tuple[int, int, int, int, str]]:
        num = self.read_value()
        result = []
        for _ in range(num):
            nameID = self.read_value()
            platID = self.read_value()
            encID = self.read_value()
            langID = self.read_value()
            name_length = self.read_value()
            name_codes = [self.read_value() for _ in range(name_length)]
            name = ""
            for c in name_codes:
                try:
                    char = chr(c)
                    # Fix platform-specific encodings for Mac
                    if platID == 1 and encID == 0:
                        # TODO: Remove default arguments when Python < 3.11 is dropped
                        char = c.to_bytes(length=1, byteorder="big").decode("macroman")
                except ValueError:
                    char = "\ufeff"
                name += char
            result.append((nameID, platID, encID, langID, name))

        return result


class OpenTypeStringParser(BaseParser):
    """
    A parser that reads data as a strings and returns it as a list.
    """

    @staticmethod
    def build_fea_dict(data: list[str]) -> FeaturesDict:
        """
        Parse the OpenType feature code into a structure that is more manageable. This
        is not used in vfbLib, but may be used in code that wants to read OpenType
        features from a VFB.

        Args:
            data (list[str]): The OpenType feature code (AFDKO syntax) as a list of
                lines.

        Returns:
            dict[str, list[dict]]: The feature code split into prefix and actual
                features.
        """
        fea = FeaturesDict(prefix=[], features=[])
        prefix: list[str] = []
        feature: list[str] = []
        tag = ""
        in_prefix = True
        for line in data:
            if "#" in line:
                code, _ = line.split("#", 1)
            else:
                code = line

            if s := search(r"\s*feature\s*([a-z0-9]{4})\s*\{", code):
                # New feature

                if in_prefix:
                    # Flush the prefix code
                    fea["prefix"] = prefix
                    in_prefix = False
                else:
                    # Flush the previous feature
                    fea["features"].append({"tag": tag, "code": feature})

                tag = s.groups()[0]
                feature = [line]
            elif in_prefix:
                prefix.append(line)
            else:
                feature.append(line)

        # Flush the last feature
        fea["features"].append({"tag": tag, "code": feature})

        return fea

    def _parse(self) -> list[str]:
        s = self.read_str_all()
        # Filter more than 2 consecutive empty lines
        lines = []
        c = 0
        for line in s.splitlines():
            if line.strip():
                c = 0
                lines.append(line)
            else:
                if c < 2:
                    lines.append(line)
                c += 1

        if len(lines) > 1:
            # Remove empty lines at the end, except one
            while not lines[-1]:
                lines.pop()
        lines.append("")
        return lines


class StringParser(BaseParser):
    """
    A parser that reads data as strings.
    """

    def _parse(self):
        return self.read_str_all().strip("\00")



import logging

from vfbLib.parsers.base import BaseParser
from vfbLib.typing import BackgroundImageDict, BitmapDataDict, GlyphBitmapDict

logger = logging.getLogger(__name__)


class BaseBitmapParser(BaseParser):
    def _read_bitmap(self, datalen: int) -> list[list[int]]:
        # Read the bitmap data from the stream.
        # The returned value is in the run-length encoded representation.
        chunks: list[list[int]] = []
        bytes_remaining = datalen
        while bytes_remaining > 0:
            bytes_remaining -= 1
            # The next value specifies how many bytes to read and how to repeat them
            v = self.read_int8()
            chunk = [v]
            if v >= 0:
                # The next n bytes are actually stored in the data
                n = v + 1
                if n > bytes_remaining:
                    break

                for _ in range(n):
                    i = self.read_uint8()
                    chunk.append(i)
                bytes_remaining -= n
            else:
                # The next n bytes are identical, read only one byte and repeat it
                n = -v + 1
                if bytes_remaining < 1:
                    break

                i = self.read_uint8()
                chunk.append(i)
                bytes_remaining -= 1
            chunks.append(chunk)
        return chunks

    def _decode_bitmap(self, chunks: list[list[int]]) -> list[int]:
        # Decode from the run-length-encoded format
        decoded: list[int] = []
        for chunk in chunks:
            bytes_spec = chunk.pop(0)
            if bytes_spec >= 0:
                decoded.extend(chunk)
            else:
                num_values = -bytes_spec + 1
                decoded.extend(chunk * num_values)
        return decoded

    def _parse_bitmap_data(
        self,
        w: int,
        datalen: int,
        origin_top: bool = False,
        preview: bool = True,
    ) -> BitmapDataDict:
        rle_data = self._read_bitmap(datalen)
        data = self._decode_bitmap(rle_data)

        # For the "windows" platform, the bitmap must be inverted.
        if self.vfb is not None and self.vfb.writer_platform == "windows":
            data = [~b + 0x100 for b in data]

        # Split list into rows
        bytes_per_row = ((w + 15) // 16) * 2
        # fmt:off
        nice = [
            data[j : j + bytes_per_row]  # noqa: E203
            for j in range(0, len(data), bytes_per_row)
        ]
        # fmt:on

        bitmap = BitmapDataDict(data=nice)

        # Generate an ASCII art preview
        if preview:
            rows = self._get_preview(bitmap["data"], w)
            if origin_top:
                bitmap["preview"] = [r for r in reversed(rows)]
            else:
                bitmap["preview"] = rows

        return bitmap

    def _get_preview(self, data: list[list[int]], payload_width: int) -> list[str]:
        # Add a block graphics bitmap preview to the decompiled data
        rows = []
        for row in data:
            preview_row = "".join(
                [f"{r:08b}".replace("0", " ▕").replace("1", "█▉") for r in row]
            )
            # Shorten row to the specified number of columns w
            preview_row = preview_row[: payload_width * 2]

            rows.append(preview_row)

        return rows


class BackgroundBitmapParser(BaseBitmapParser):
    def _parse(self) -> BackgroundImageDict:
        bitmap = BackgroundImageDict(
            origin=(0, 0),
            size_units=(0, 0),
            size_pixels=(0, 0),
            bitmap=BitmapDataDict(data=[]),
        )
        bitmap["origin"] = (self.read_value(), self.read_value())
        bitmap["size_units"] = (
            self.read_value(signed=False),
            self.read_value(signed=False),
        )
        w = self.read_value(signed=False)
        h = self.read_value(signed=False)
        bitmap["size_pixels"] = (w, h)
        datalen = self.read_value(signed=False)
        if not hasattr(self, "preview"):
            self.preview = True
        bitmap["bitmap"] = self._parse_bitmap_data(w, datalen, preview=self.preview)
        return bitmap


class GlyphBitmapsParser(BaseBitmapParser):
    def _parse(self) -> list[GlyphBitmapDict]:
        bitmaps: list[GlyphBitmapDict] = []
        num_bitmaps = self.read_value(signed=False)
        for _ in range(num_bitmaps):
            bitmap = GlyphBitmapDict(
                ppm=0,
                origin=(0, 0),
                adv=(0, 0),
                size_pixels=(0, 0),
                bitmap=BitmapDataDict(data=[]),
            )
            bitmap["ppm"] = self.read_value(signed=False)
            bitmap["origin"] = (self.read_value(), self.read_value())
            bitmap["adv"] = (
                self.read_value(signed=False),
                self.read_value(signed=False),
            )
            w = self.read_value(signed=False)
            h = self.read_value(signed=False)
            bitmap["size_pixels"] = (w, h)
            datalen = self.read_value()
            bitmap["bitmap"] = self._parse_bitmap_data(w, datalen)
            bitmaps.append(bitmap)
        return bitmaps



from typing import TYPE_CHECKING

from vfbLib import ExpandKernOptions, TTAutoHintOptions, export_options, font_options
from vfbLib.compilers.base import BaseCompiler
from vfbLib.helpers import intListToBinary

if TYPE_CHECKING:
    from vfbLib.typing import FontOptionsDict


class ExportOptionsCompiler(BaseCompiler):
    def _compile(self, data: list[str]) -> None:
        rev = {v: k for k, v in export_options.items()}
        options = [rev[k] for k in data]
        self.write_uint16(intListToBinary(options))


class FontOptionsCompiler(BaseCompiler):
    def _compile(self, data: "FontOptionsDict") -> None:
        rev = {v: k for k, v in font_options.items()}
        for k, v in data.items():
            raw_key = rev.get(k)
            assert raw_key is not None
            match k:
                case "auto_hinting_h_ratio" | "auto_hinting_v_ratio":
                    assert isinstance(v, float)
                    v = round(v * 10_000)
                case "autohinting_options":
                    assert isinstance(v, dict)
                    packed = v.get("single_link_attachment_precision", 7)
                    for subkey in (
                        "generate_triple_hints",
                        "generate_delta_instructions",
                        "direct_links_to_center_of_the_glyph_where_possible",
                        "interpolate_positions_of_cusp_points",
                        "interpolate_positions_of_double_links",
                        "add_link_to_rsb",
                    ):
                        packed += v.get(subkey, 0) * getattr(TTAutoHintOptions, subkey)
                    v = packed
                case "expand_kern_flags":
                    assert isinstance(v, dict)
                    packed = 0
                    for subkey in (
                        "limit_action",
                        "limit_codepage",
                        "limit_cmap_10",
                        "limit_font_window",
                        "limit_count",
                        "limit_keep",
                        "apply_to_assistance",
                    ):
                        packed += v.get(subkey, 0) * getattr(ExpandKernOptions, subkey)
                    v = packed
                case _:
                    pass
            self.write_uint8(raw_key)
            assert isinstance(v, int)
            self.write_value(v)
        self.write_uint8(0x64)



import logging
from functools import cached_property
from io import BytesIO
from struct import pack
from typing import TYPE_CHECKING

from vfbLib.compilers.base import BaseCompiler
from vfbLib.constants import parser_classes
from vfbLib.enum import F
from vfbLib.helpers import hexStr, int32_size
from vfbLib.parsers.base import BaseParser, StreamReader
from vfbLib.typing import EntryDict

if TYPE_CHECKING:
    from io import BufferedIOBase

    from vfbLib.typing import EntryDecompiled
    from vfbLib.vfb.vfb import Vfb


logger = logging.getLogger(__name__)


FALLBACK_PARSER = BaseParser


class VfbEntry(StreamReader):
    def __init__(
        self,
        parent: "Vfb",
        parser: type[BaseParser] | None = None,
        compiler: type[BaseCompiler] | None = None,
        eid: int | None = None,
    ) -> None:
        # The parent object, Vfb
        self.vfb = parent
        # The original or decompiled data
        self._data: "bytes | EntryDecompiled | None" = None
        # Temporary data for additional master, must be merged when compiling
        self.temp_masters: list[list] | None = None
        self.parser = None
        self.compiler = None
        # The numeric and human-readable key of the entry, also sets parser & compiler
        self.id = eid
        # The parser which can convert data to decompiled.
        # Use the arg to override the parser looked up in the id setter only
        if parser is not None:
            logger.warning(f"Overriding parser {self.parser} with {parser}")
            self.parser = parser
        # The compiler which can convert the decompiled to compiled data.
        # Use the arg to override the compiler looked up in the id setter only
        if compiler is not None:
            logger.warning(f"Overriding compiler {self.compiler} with {compiler}")
            self.compiler = compiler

    def __repr__(self) -> str:
        return (
            f"<VfbEntry {self.id} ({self.key}), parser: {self.parser}, "
            f"compiler: {self.compiler}>"
        )

    @cached_property
    def header(self) -> bytes:
        """The entry header.

        Returns:
            bytes: The data of the current entry header.
        """
        if self.id is None:
            logger.error(
                "You need to set VfbEntry.id before accessing its header, "
                "usually by reading the entry from the input stream."
            )
            raise ValueError

        header = BytesIO()
        if self.size > 0xFFFF:
            entry_id = self.id | 0x8000
        else:
            entry_id = self.id
        header.write(pack("<H", entry_id))
        if self.size > 0xFFFF:
            header.write(pack("<I", self.size))
        else:
            header.write(pack("<H", self.size))
        return header.getvalue()

    @property
    def decompiled(self) -> "EntryDecompiled | None":
        """
        Deprecated, use VfbEntry.data
        """
        import warnings

        warnings.warn(
            "VfbEntry.decompiled is deprecated, use VfbEntry.data instead",
            DeprecationWarning,
        )
        if isinstance(self._data, bytes):
            return None

        return self._data

    @property
    def size(self) -> int:
        """The size of the compiled data.

        Returns:
            int: The size of the current compiled data.
        """
        if self.data is None:
            return 0

        if isinstance(self.data, bytes):
            return len(self.data)

        raise RuntimeError

    @property
    def data(self) -> "bytes | EntryDecompiled | None":
        return self._data

    @data.setter
    def data(self, value: "bytes | EntryDecompiled | None") -> None:
        self._data = value

    @property
    def id(self) -> int | None:
        return self._id

    @id.setter
    def id(self, value: int | None) -> None:
        self._id = value
        self.key = str(self._id)
        self.compiler = None
        if self._id is None:
            self.parser = None
        else:
            entry_info = parser_classes.get(self._id)
            if entry_info is None:
                logger.warning(f"Could not find entry info for ID {self._id}")
                logger.warning(
                    f"Falling back to parser {FALLBACK_PARSER} for ID {self._id}"
                )
                self.parser = FALLBACK_PARSER
            else:
                self.key, self.parser, self.compiler = entry_info
                if self.parser is not None:
                    self.parser.encoding = self.vfb.encoding

    def _read_entry(self) -> int:
        """
        Read an entry from the stream and return its data size.
        """
        raw_id = self.read_uint16()
        self.id = raw_id & ~0x8000

        if raw_id & 0x8000:
            # Uses uint32 for data length
            num_bytes = self.read_uint32()
        else:
            # Uses uint16 for data length
            num_bytes = self.read_uint16()

        return num_bytes

    def as_dict(self, minimize=True) -> EntryDict:
        d = EntryDict(key=str(self.key))
        if isinstance(self.data, bytes):
            d["size"] = self.size
            d["decompiled"] = hexStr(self.data)
        else:
            d["decompiled"] = self.data
        if not minimize:
            if self.parser is not None:
                d["parser"] = self.parser.__name__
            if self.compiler is not None:
                d["compiler"] = self.compiler.__name__
        return d

    def compile(self) -> bool:
        """
        Compile the entry. The result is stored in VfbEntry.data.

        Returns:
            bool: Whether compilation was successful.
        """
        if isinstance(self.data, bytes):
            # Is already compiled
            return True

        if self.compiler is None:
            logger.error(
                f"Compiling '{self.id}' is not supported yet in {self} "
                f"Data: {self.data}"
            )
            return False

        self.merge_masters_data()

        self.data = self.compiler().compile(self.data, vfb=self.vfb)

        # TODO: Return False here if compilation has failed. How to tell?

        return True

    def decompile(self) -> None:
        """
        Decompile the entry. The result is stored in VfbEntry.data.
        """
        if self.parser is None:
            raise ValueError(f"No parser is specified for entry type {self.id}")

        if self.data is None:
            raise ValueError

        if not isinstance(self.data, bytes):
            # Already decompiled
            return

        byte_data = self.data

        try:
            self.data = self.parser().parse(BytesIO(byte_data), self.size, self.vfb)
        except:  # noqa: E722
            logger.error(f"Parse error for data: {self.key}; {hexStr(byte_data)}")
            logger.error(f"Parser class: {self.parser.__name__}")
            # self.data = byte_data
            self.error = "ERROR"  # TODO: Include traceback
            self.vfb.any_errors |= True
            raise

    def merge_masters_data(self) -> None:
        """
        Merge any temporary masters data into the main decompiled structure. Such data
        can be added as the result of drawing with a PointPen into a multiple master
        Vfb.
        """
        if self.temp_masters is None:
            return

        if self.vfb.num_masters == 1:
            return

        if self.compiler is None:
            return

        self.compiler.merge(self.temp_masters, self.data)

    def read(self, stream: "BufferedIOBase") -> None:
        """
        Read the entry from the stream without decompiling the data.
        """
        self.stream = stream
        size = self._read_entry()

        # Special cases, probably remnants from FL3
        if self.id == F.MMKernPair:
            # We can't trust the size given in the "length" value of the entry.
            # In this case it can be calculated from the number of masters:
            size = 2 * int32_size + self.vfb.num_masters * 2

        self.data = self.stream.read(size)



from typing import NotRequired, TypedDict

from vfbLib.typing import HintTuple


class TUfoTTZoneDict(TypedDict):
    position: int
    top: bool
    width: int
    # Zone index, shift; index must be str to be saved in lib
    delta: NotRequired[dict[str, int]]


UfoPoint = tuple[int, int]
UfoComponent = tuple[str, tuple[float, float, float, float, int, int]]
UfoGroups = dict[str, list[str]]  # name, glyphs
UfoMasterKerning = dict[tuple[str, str], int]  # Lstr, Rstr, value
UfoMMKerning = dict[tuple[str, int], list[int]]  # Lstr, Rid, master values
UfoSegment = tuple[str | None, bool, str | None, UfoPoint]
UfoContour = list[UfoSegment]
TUfoTTZonesDict = dict[str, TUfoTTZoneDict]


class TUfoRawStemDict(TypedDict):
    name: str
    round: dict[str, int]
    value: int


class TUfoRawStemsDict(TypedDict):
    ttStemsH: list[TUfoRawStemDict]
    ttStemsV: list[TUfoRawStemDict]


class TUfoGaspRecDict(TypedDict):
    rangeMaxPPEM: int
    rangeGaspBehavior: list[int]


class TUfoStemDict(TypedDict):
    horizontal: bool
    name: str
    round: dict[str, int]
    width: int


class TUfoStemsDict(TypedDict):
    ttStemsH: list[TUfoStemDict]
    ttStemsV: list[TUfoStemDict]


class TUfoStemPPMDict(TypedDict):
    stem: int
    round: dict[int, int]


class TUfoStemPPMsDict(TypedDict):
    ttStemsH: list[TUfoStemPPMDict]
    ttStemsV: list[TUfoStemPPMDict]


class UfoGuide(TypedDict):
    angle: float | int
    color: NotRequired[str]
    name: NotRequired[str]
    x: int
    y: int
    _direction: NotRequired[str]


class HintSet(TypedDict):
    pointTag: str
    stems: list[HintTuple]


class UfoHintSet(TypedDict):
    pointTag: str
    stems: list[str]


class UfoHintingV2(TypedDict):
    flexList: NotRequired[list[str]]
    hintSetList: NotRequired[list[UfoHintSet]]
    id: NotRequired[str]



import logging

from vfbLib.helpers import hexStr
from vfbLib.parsers.base import BaseParser

logger = logging.getLogger(__name__)


class BinaryTableParser(BaseParser):
    def _parse(self) -> dict[str, str]:
        tag = self.read_str(4)
        data = hexStr(self.stream.read())
        return {"tag": tag, "data": data}



import logging

from vfbLib.parsers.base import BaseParser

logger = logging.getLogger(__name__)


class MMKernPairParser(BaseParser):
    """
    A parser that reads the special-cased MM Kern Pair entry, which is apparently only
    written by FontLab 3.
    """

    def _parse(self) -> dict[str, int | list[int]]:
        assert self.vfb is not None
        index_1 = self.read_uint32()
        index_2 = self.read_uint32()
        pair: dict[str, int | list[int]] = {"l": index_1, "r": index_2}
        values: list[int] = []
        for _ in range(self.vfb.num_masters):
            values.append(self.read_int16())
        pair["values"] = values
        return pair



import codecs
from argparse import ArgumentParser
from copy import deepcopy
from pathlib import Path
from sys import exit
from typing import Any

import orjson

from vfbLib.ufo.glyph import IndexVfbToUfoGlyph
from vfbLib.ufo.tth import TTGlyphHints, transform_stem_rounds
from vfbLib.vfb.vfb import Vfb


def vfb2tth():
    """
    The command line interface for exporting TrueType hinting in FontLab's high-level
    format to JSON, TOML, or YAML.
    """
    parser = ArgumentParser(
        description="vfb2tth Converter\nCopyright (c) 2024 by LucasFonts"
    )
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        default="json",
        help="The output format: json (default), toml, or yaml",
    )
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        nargs=1,
        help="output folder",
    )
    parser.add_argument(
        "inputpath",
        type=str,
        nargs=1,
        help="input file path (.vfb)",
    )
    args = parser.parse_args()
    if args:
        vfb_path = Path(args.inputpath[0])
        print(parser.description)
        print(f"Reading file {vfb_path} ...")
        vfb = Vfb(
            vfb_path,
            only_header=False,
            minimal=True,
            unicode_strings=True,
        )
        suffix = f".tth.{args.format}"
        if args.path:
            out_path = (Path(args.path[0]) / vfb_path.name).with_suffix(suffix)
        else:
            out_path = vfb_path.with_suffix(suffix)
        data = extract_truetype_hinting(vfb)
        if args.format == "toml":
            import tomli_w

            with open(out_path, "wb") as f:
                tomli_w.dump(data, f)
        elif args.format == "yaml":
            import yaml

            with codecs.open(str(out_path), "wb", "utf-8") as f:
                yaml.dump(data, f, sort_keys=True, indent=2)
        else:
            with open(str(out_path), "wb") as f:
                f.write(
                    orjson.dumps(
                        data,
                        option=orjson.OPT_INDENT_2
                        | orjson.OPT_NON_STR_KEYS
                        | orjson.OPT_SORT_KEYS,
                    )
                )

        if vfb.any_errors:
            exit("There were decompilation errors.")
    else:
        parser.print_help()
        exit(2)


def extract_truetype_hinting(vfb: Vfb) -> dict[str, Any]:
    """
    Collect the relevant entries from the VFB and extract the hinting information by
    calling a specialized function for each data type. The hinting information is
    returned as a dict which can be serialized in the desired format.

    Args:
        vfb (Vfb): The input VFB.

    Returns:
        dict[str, Any]: The extracted TrueType hinting information.
    """
    font: dict[str, Any] = {}
    glyphs: dict[str, list[dict]] = {}
    zone_names: dict[str, dict] = {"ttZonesT": {}, "ttZonesB": {}}
    d: dict[str, Any] = {"font": font, "glyphs": glyphs}
    stem_round: dict[str, dict] = {"ttStemsV": {}, "ttStemsH": {}}
    for entry in vfb.entries:
        key = entry.key
        if key == "gasp":
            entry.decompile()
            assert isinstance(entry.data, list)
            font["gasp"] = {
                str(gasp_range["maxPpem"]): gasp_range["flags"]
                for gasp_range in entry.data
            }

        elif key == "ttinfo":
            entry.decompile()
            assert isinstance(entry.data, dict)
            for k in ("head_lowest_rec_ppem", "head_units_per_em"):
                if k in entry.data:
                    font[k] = entry.data[k]

        elif key == "TrueType Stem PPEMs":
            entry.decompile()
            assert isinstance(entry.data, dict)
            extract_tt_stem_ppems(entry.data, stem_round)

        elif key == "TrueType Stems":
            entry.decompile()
            assert isinstance(entry.data, dict)
            extract_tt_stems(entry.data, font)

        elif key == "TrueType Stem PPEMs 1":
            entry.decompile()
            assert isinstance(entry.data, dict)
            extract_tt_stem_ppem_1(entry.data, stem_round)

        elif key == "TrueType Zones":
            entry.decompile()
            assert isinstance(entry.data, dict)
            extract_tt_zones(entry.data, font, zone_names)

        elif key == "stemsnaplimit":
            entry.decompile()
            assert isinstance(entry.data, int)
            font["stemsnaplimit"] = entry.data

        elif key == "zoneppm":
            entry.decompile()
            assert isinstance(entry.data, int)
            font["zoneppm"] = entry.data

        elif key == "codeppm":
            entry.decompile()
            assert isinstance(entry.data, int)
            font["codeppm"] = entry.data

        elif key == "TrueType Zone Deltas":
            entry.decompile()
            assert isinstance(entry.data, dict)
            extract_tt_zone_deltas(entry.data, font)

        elif key == "Glyph":
            entry.decompile()
            assert isinstance(entry.data, dict)
            extract_glyph_hints(entry.data, glyphs, font, zone_names)

    merge_stem_information(stem_round, font)
    merge_zone_information(font)

    return d


def extract_glyph_hints(
    data: dict, target: dict, font_hints: dict, zone_names: dict
) -> None:
    """
    Extract the glyph hints from the supplied decompiled glyph entry `data`.

    Args:
        data (dict): The decompiled glyph entry data.
        target (dict): The dict to which to add the extracted glyph hints.
        font_hints (dict): Font-level hinting data. Used to map stem indices to names.
        zone_names (dict): Zone hinting data. Used to map zone indices to names.
    """
    if tth := data.get("tth"):
        ttg = TTGlyphHints(
            IndexVfbToUfoGlyph(),
            data=tth,
            zone_names=zone_names,
            stems=font_hints["stems"],
        )
        tr = ttg.get_tt_glyph_hints()
        target[data["name"]] = tr


def extract_tt_stem_ppem_1(data: dict, target: dict) -> None:
    """
    Extract the stem pixel width information for 1 pixel from the supplied decompiled
    TrueType stem ppem1 `data`.

    Args:
        data (dict): The decompiled entry data.
        target (dict): The dict to which to add the extracted information.
    """
    for direction in ("ttStemsV", "ttStemsH"):
        if direction in data:
            direction_data = data[direction]
            for stem_data in direction_data:
                stem_index = stem_data["stem"]
                target[direction][stem_index]["1"] = stem_data["round"]["1"]


def extract_tt_stem_ppems(data: dict, target: dict) -> None:
    """
    Extract the stem pixel width information for 2–6 pixels from the supplied decompiled
    TrueType stem ppem `data`.

    Args:
        data (dict): The decompiled entry data.
        target (dict): The dict to which to add the extracted information.
    """
    for direction in ("ttStemsV", "ttStemsH"):
        if direction in data:
            direction_data = data[direction]
            for stem_data in direction_data:
                target[direction][stem_data["stem"]] = deepcopy(stem_data["round"])


def extract_tt_stems(data: dict, target: dict) -> None:
    """
    Extract the stem information from the supplied decompiled TrueType stem `data`.

    Args:
        data (dict): The decompiled entry data.
        target (dict): The dict to which to add the extracted information.
    """
    target["stems"] = {"ttStemsV": [], "ttStemsH": []}
    stem_names = set()
    for direction in ("ttStemsV", "ttStemsH"):
        for stem in data.get(direction, []):
            name = stem["name"]
            i = 1
            # Make stem names unique
            if name in stem_names:
                print(f"Duplicate stem name: {name}")
                while f"{name}#{i}" in stem_names:
                    i += 1
                name = f"{name}#{i}"
            stem_names |= {name}
            target["stems"][direction].append(
                {
                    "name": name,
                    "round": deepcopy(stem["round"]),
                    "width": stem["value"],
                    "horizontal": direction == "ttStemsV",
                }
            )


def extract_tt_zone_deltas(data: dict, target: dict):
    """
    Extract the zone delta information from the supplied decompiled TrueType zone deltas
    `data`.

    Args:
        data (dict): The decompiled entry data.
        target (dict): The dict to which to add the extracted information.
    """
    target["zone_deltas"] = deepcopy(data)


def extract_tt_zones(data: dict, target: dict, zone_names: dict) -> None:
    """
    Extract the zone information from the supplied decompiled TrueType zone `data`.

    Args:
        data (dict): The decompiled entry data.
        target (dict): The dict to which to add the extracted information.
        zone_names (dict): The dict to which the zone names will be added.
    """
    target["zones"] = {"ttZonesT": [], "ttZonesB": []}
    zone_names_global = set()
    for side in ("ttZonesT", "ttZonesB"):
        for zone_index, zone in enumerate(data.get(side, [])):
            name = zone["name"]
            i = 1
            # Make zone names unique
            if name in zone_names_global:
                print(f"Duplicate zone name: {name}")
                while f"{name}#{i}" in zone_names_global:
                    i += 1
                name = f"{name}#{i}"
            zone_names_global |= {name}
            zone_names[side][zone_index] = name
            target["zones"][side].append(
                {
                    "name": name,
                    "position": zone["position"],
                    "top": side == "ttZonesT",
                    "width": zone["value"],
                }
            )


def merge_stem_information(data: dict, font: dict) -> None:
    """
    Merge stem information from various fields in the `data` dict into the font dict in
    the structure expected by the export format.

    Args:
        data (dict): The input dict.
        font (dict): The output dict.
    """
    stems_dict = {}
    for direction in ("ttStemsV", "ttStemsH"):
        if direction in data:
            direction_data = data[direction]
            for stem_index, rounding in direction_data.items():
                stem = font["stems"][direction][stem_index]
                stem["round"].update(rounding)
                # Sort the rounding dict by value (px)
                rd = transform_stem_rounds(stem["round"], stem["name"])
                srtl = sorted(rd.items(), key=lambda item: item[1])
                srd = {k: v for k, v in srtl}
                stem["round"] = srd
                stems_dict[stem["name"]] = deepcopy(stem)
                del stems_dict[stem["name"]]["name"]
    if stems_dict:
        font["stems"] = stems_dict
    else:
        del font["stems"]


def merge_zone_information(font: dict) -> None:
    """
    Merge zone information from various fields into the structure expected by the export
    format.

    Args:
        font (dict): _description_
    """
    all_zones = []
    for side in ("ttZonesB", "ttZonesT"):
        for zone in font["zones"].get(side, []):
            all_zones.append(zone)
    for zone_index, zone_delta in font.get("zone_deltas", {}).items():
        zone = all_zones[zone_index]
        # Sort and convert delta ppm to str
        zone["deltas"] = {str(k): v for k, v in sorted(zone_delta.items())}

    zones_dict = {}
    for zone in all_zones:
        zones_dict[zone["name"]] = deepcopy(zone)
        del zones_dict[zone["name"]]["name"]
    if zones_dict:
        font["zones"] = zones_dict
    else:
        del font["zones"]
    del font["zone_deltas"]



from vfbLib.helpers import binaryToIntList
from vfbLib.parsers.base import BaseParser
from vfbLib.typing import BBoxDict, GlyphHintingOptionsDict, PSInfoDict

global_options = (
    ("generate_flex", 0),  # Automatically generate Flex hints in T1 and OT
)

glyph_options = (
    ("hint_replacement", 29),  # Hint Replacement
    ("horizontal_3_stem", 30),  # Generate horizontal 3-stem
    ("vertical_3_stem", 31),  # Generate vertical 3-stem
)


class PostScriptInfoParser(BaseParser):
    """
    A parser that reads data into a dict of PostScript font info.
    """

    def _parse(self) -> PSInfoDict:
        values: PSInfoDict = {}
        values["font_matrix"] = self.read_doubles(6)
        values["force_bold"] = self.read_int32()
        values["blue_values"] = [self.read_int32() for _ in range(14)]
        values["other_blues"] = [self.read_int32() for _ in range(10)]
        values["family_blues"] = [self.read_int32() for _ in range(14)]
        values["family_other_blues"] = [self.read_int32() for _ in range(10)]
        values["blue_scale"] = self.read_double()
        values["blue_shift"] = self.read_uint32()
        values["blue_fuzz"] = self.read_uint32()
        values["std_hw"] = self.read_uint32()
        values["std_vw"] = self.read_uint32()
        values["stem_snap_h"] = [self.read_uint32() for _ in range(12)]
        values["stem_snap_v"] = [self.read_uint32() for _ in range(12)]
        # The bounding box values only get updated during some actions, e.g.
        # going into PS hinting mode
        xMin, yMin, xMax, yMax = [self.read_int16() for _ in range(4)]
        values["bounding_box"] = BBoxDict(xMin=xMin, yMin=yMin, xMax=xMax, yMax=yMax)
        values["adv_width_min"] = self.read_int32()
        values["adv_width_max"] = self.read_int32()
        values["adv_width_avg"] = self.read_int32()
        values["ascender"] = self.read_int32()
        values["descender"] = self.read_int32()
        values["x_height"] = self.read_int32()
        values["cap_height"] = self.read_int32()
        return values


class PostScriptGlobalHintingOptionsParser(BaseParser):
    """
    A parser that reads data as a bitfield describing a glyph's PostScript hinting
    options.
    """

    def _parse(self):
        bits = set(binaryToIntList(self.read_uint16()))
        d = {}
        for k, bit in global_options:
            if bit in bits:
                bits.discard(bit)
                d[k] = 1

        d["other"] = sorted(bits)
        return d


class PostScriptGlyphHintingOptionsParser(BaseParser):
    """
    A parser that reads data as a bitfield describing a glyph's PostScript hinting
    options.
    """

    def _parse(self) -> GlyphHintingOptionsDict:
        bits = set(binaryToIntList(self.read_uint32()))
        d: GlyphHintingOptionsDict = {}
        for k, bit in glyph_options:
            if bit in bits:
                bits.discard(bit)
                d[k] = 1

        d["other"] = sorted(bits)
        return d



from typing import TYPE_CHECKING

from vfbLib.helpers import int8_size, int32_size

if TYPE_CHECKING:
    from io import BufferedIOBase


def read_value(stream: "BufferedIOBase", signed=True) -> int:
    """
    Read an encoded value from the stream, decode it to integer and return it.

    The storage format in the stream of the "encoded value" is the same as described in
    the Type 1 specification, p. 48, "Charstring Number Encoding"
    <https://adobe-type-tools.github.io/font-tech-notes/pdfs/T1_SPEC.pdf>.

    Args:
        stream (BufferedIOBase): The input stream.
        signed (bool, optional): Whether to interpret the value as a signed integer.
            Defaults to True.

    Raises:
        EOFError: When a 0 byte is encountered while reading the stream.
        ValueError: When the byte that was read from the stream is < 0x20.
        ValueError: When the byte that was read from the stream has any other undefined
            value.

    Returns:
        int: The decoded value.
    """
    val = int.from_bytes(stream.read(int8_size), byteorder="little")
    if val == 0:
        raise EOFError

    if val < 0x20:
        raise ValueError

    elif val < 0xF7:
        # -107 to 107, represented by 1 byte
        return val - 0x8B

    elif val <= 0xFA:
        # 108 to 1131, represented by 2 bytes
        val2 = int.from_bytes(stream.read(int8_size), byteorder="little")
        # val - 0x8B + (val - 0xF7) * 0xFF + val2
        return 0x100 * val - 0xF694 + val2

    elif val <= 0xFE:
        # -108 to -1131, represented by 2 bytes
        val2 = int.from_bytes(stream.read(int8_size), byteorder="little")
        # 0x8F - val - (val - 0xFB) * 0xFF - val2
        return -0x100 * val + 0xFA94 - val2

    elif val == 0xFF:
        # 32 bit integer follows
        # FIXME: The Type1 spec says that it is always a signed int.
        decoded = int.from_bytes(
            stream.read(int32_size), byteorder="big", signed=signed
        )
        return decoded

    raise ValueError



from io import BytesIO
from struct import pack
from typing import TYPE_CHECKING

from vfbLib.compilers.value import write_value, write_value_long
from vfbLib.helpers import deHexStr, hexStr, int8_size, int16_size, int32_size

if TYPE_CHECKING:
    from io import BufferedIOBase
    from typing import Any, Iterable

    from vfbLib.typing import (
        KerningClassFlagDict,
        MappingModeDict,
        MetricsClassFlagDict,
    )
    from vfbLib.vfb.vfb import Vfb


# Compilers for VFB entries


class StreamWriter:
    """
    Base compiler class that writes values to the output stream.
    This is the parent class for the general BaseCompiler, from which all other
    compilers inherit, but it may be subclassed directly if more flexibility is needed.
    """

    def __init__(self) -> None:
        self.encoding = "cp1252"
        self.stream: "BufferedIOBase" = BytesIO()

    def write_bytes(self, value: bytes) -> None:
        """
        Write binary data to the stream.

        Args:
            value (bytes): The data.
        """
        self.stream.write(value)

    def write_double(self, value: float) -> None:
        """
        Write a double-precision float value to the stream.

        Args:
            value (float): The float value to write.
        """
        self.stream.write(pack("d", value))

    def write_doubles(self, values: "Iterable[float]") -> None:
        """
        Write several double-precision floats to the stream.

        Args:
            values (Iterable[float]): The sequence of double-precision floats to write.
        """
        for f in values:
            self.write_double(f)

    def write_int8(self, value: int) -> None:
        """
        Write a signed 8-bit integer to the stream.

        Args:
            value (int): The integer value to write.
        """
        self.stream.write(value.to_bytes(int8_size, byteorder="little", signed=True))

    def write_int16(self, value: int) -> None:
        """
        Write a signed 16-bit integer to the stream.

        Args:
            value (int): The integer value to write.
        """
        self.stream.write(value.to_bytes(int16_size, byteorder="little", signed=True))

    def write_int32(self, value: int) -> None:
        """
        Write a signed 32-bit integer to the stream.

        Args:
            value (int): The integer value to write.
        """
        self.stream.write(value.to_bytes(int32_size, byteorder="little", signed=True))

    def write_str(self, value: str | None, pad: int = 0) -> None:
        """
        Write a string to the stream. The string is encoded with
        `StreamWriter.encoding`.

        Args:
            value (str | None): The string to write.
            pad (int, optional): Pad the string to the given length with null bytes.
                Defaults to 0.
        """
        if value is None:
            value = ""
        enc = value.encode(self.encoding)
        if len(enc) < pad:
            enc += b"\00" * (pad - len(enc))
        self.stream.write(enc)

    def write_str_with_len(self, value: str | None) -> None:
        """
        Write the length of a string to the stream, then the string itself. The string
        is encoded with `StreamWriter.encoding`.

        Args:
            value (str | None): The string to write.
        """
        if value is None:
            value = ""
        enc = value.encode(self.encoding)
        self.write_value(len(enc))
        self.stream.write(enc)

    def write_uint8(self, value: int) -> None:
        """
        Write a uint8 value to the stream.

        Args:
            value (int): The integer to write.
        """
        self.stream.write(value.to_bytes(int8_size, byteorder="little", signed=False))

    def write_uint16(self, value: int) -> None:
        """
        Write a uint16 value to the stream.

        Args:
            value (int): The integer to write.
        """
        self.stream.write(value.to_bytes(int16_size, byteorder="little", signed=False))

    def write_uint32(self, value: int) -> None:
        """
        Write an unsigned 32-bit integer to the stream.

        Args:
            value (int): The integer value to write.
        """
        self.stream.write(value.to_bytes(int32_size, byteorder="little", signed=False))

    def write_value(self, value: int, shortest=True, signed: bool = True) -> None:
        """
        Encode and write an int value to the stream. Optionally don't apply the length
        encoding optimization.

        Args:
            value (int): The value to write to the stream.
            shortest (bool, optional): Whether to write the shortest possible
                representation. Defaults to True.
        """
        if shortest:
            write_value(value, self.stream, signed)
        else:
            write_value_long(value, self.stream, signed)


class BaseCompiler(StreamWriter):
    """
    Base class to compile vfb data.
    """

    def compile(self, data: "Any", vfb: "Vfb | None" = None) -> bytes:
        """
        Compile the JSON-like main data structure and return the compiled binary data.

        The actual compilation is done by calling the `_compile` method, which must be
        implemented for all specialized compiler subclasses.

        Args:
            data (Any): The main data structure.
            vfb (int, optional): The Vfb that is calling the compiler.

        Returns:
            bytes: The compiled binary data.
        """
        self.vfb = vfb
        self.stream = BytesIO()
        self._compile(data)
        return self.stream.getvalue()

    def compile_hex(self, data: "Any", vfb: "Vfb | None" = None) -> str:
        """
        Compile the data given into a hex string format, e.g. "8c 8d 89 8b". Used for
        testing.

        Args:
            data (Any): The input data
            vfb (int, optional): The Vfb that is calling the compiler.

        Returns:
            str: The hex string
        """
        b = self.compile(data, vfb)
        return hexStr(b)

    def _compile(self, data: "Any") -> None:
        raise NotImplementedError

    @classmethod
    def merge(cls, masters_data: "list[Any]", data: "Any") -> None:
        """
        Merge the data of additional masters into the main data structure. This operates
        on the uncompiled JSON-like data structure.

        Args:
            masters_data (list[Any]): The additional masters data as a list with one
                entry per master.
            data (Any): The main data structure.
        """
        # Must be implemented for compilers that need it, e.g. the GlyphCompiler.
        pass


class EncodedValueListCompiler(BaseCompiler):
    def _compile(self, data: "Any") -> None:
        for value in data:
            self.write_value(value)


class EncodedValueListWithCountCompiler(BaseCompiler):
    def _compile(self, data: list[int]) -> None:
        self.write_value(len(data))
        for value in data:
            self.write_value(value)


class GlyphEncodingCompiler(BaseCompiler):
    def _compile(self, data: "Any") -> None:
        """
        Compile the data into the format used by a glyph encoding entry.

        Args:
            data (tuple[int, str]): A tuple containing the glyph ID and the glyph name.
        """
        gid, name = data
        self.write_uint16(gid)
        self.write_str(name)  # XXX: Does it have to be cp1252?


class HexStringCompiler(BaseCompiler):
    def _compile(self, data: str | None) -> None:
        """
        Compile the data given in hex string format to the stream as bytes.

        This can be used as a fallback for unsupported entries as long as the hex data
        is known, e.g. for end markers or constants.

        Args:
            data (str | None): The hex string data, e.g. "203955", or None.
        """
        if not data:
            return

        self.write_bytes(deHexStr(data))


mapping_modes = {
    0: "names_or_index",
    1: "unicode_ranges",
    3: "codepages",
}


class MappingModeCompiler(BaseCompiler):
    def _compile(self, data: "MappingModeDict") -> None:
        rev_map = {v: k for k, v in mapping_modes.items()}
        self.write_uint8(1)
        self.write_value(rev_map[data["mapping_mode"]])
        self.write_uint8(2)
        self.write_value(data["m2"])
        self.write_uint8(3)
        self.write_value(data["m3"])
        self.write_uint8(4)
        self.write_value(data["mapping_id"])
        self.write_uint8(0)


class OpenTypeKerningClassFlagsCompiler(BaseCompiler):
    def _compile(self, data: "KerningClassFlagDict") -> None:
        self.write_value(len(data), signed=False)
        for name, flags in data.items():
            self.write_str_with_len(name)
            flag1, flag2 = flags
            self.write_value(flag1, signed=False)
            self.write_value(flag2, signed=False)


class OpenTypeMetricsClassFlagsCompiler(BaseCompiler):
    def _compile(self, data: "MetricsClassFlagDict") -> None:
        self.write_value(len(data), signed=False)
        for name, flags in data.items():
            self.write_str_with_len(name)
            flag1, flag2, flag3 = flags
            self.write_value(flag1, signed=False)
            self.write_value(flag2, signed=False)
            self.write_value(flag3, signed=False)



from typing import TYPE_CHECKING, Any

from vfbLib import tt_settings, ttinfo_names
from vfbLib.compilers.base import BaseCompiler
from vfbLib.helpers import intListToBinary

if TYPE_CHECKING:
    from vfbLib.typing import (
        TrueTypeInfoDict,
        TTStemsDict,
        TTZoneDict,
        TTZonesDict,
        VdmxRecDict,
    )


def convert_flags_options_to_int(data: "TrueTypeInfoDict") -> int:
    value = 0
    head_flags_options = data["head_flags"]
    assert isinstance(head_flags_options, dict)
    flags_list: list[int] = head_flags_options.get("flags", [])
    value = intListToBinary(flags_list)
    options: tuple[str, ...]
    if options := tuple(head_flags_options["options"]):
        option_bits = []
        for k, v in tt_settings.items():
            if v in options:
                option_bits.append(k)

        value += intListToBinary(option_bits) << 16
    return value


class GaspCompiler(BaseCompiler):
    def _compile(self, data: Any) -> None:
        for rec in data:
            self.write_uint16(rec["maxPpem"])
            self.write_uint16(rec["flags"])


class TrueTypeInfoCompiler(BaseCompiler):
    def _write_if_exists(
        self, numkey: int, data: "TrueTypeInfoDict", signed: bool = True
    ) -> None:
        strkey = ttinfo_names[numkey]
        if strkey in data:
            self.write_uint8(numkey)
            self.write_value(data[strkey], signed=signed)

    def _compile(self, data: "TrueTypeInfoDict") -> None:
        for k in (0x33, 0x34, 0x35, 0x36, 0x37, 0x38):
            self.write_uint8(k)
            self.write_value(data[ttinfo_names[k]])

        # Convert combined flags and options dict back to a single int
        head_flags_options = convert_flags_options_to_int(data)
        self.write_uint8(0x39)
        self.write_value(head_flags_options)

        for k in (
            0x3A,
            0x3B,
            0x3C,
        ):
            self._write_if_exists(k, data)

        for k in (0x56, 0x57):
            self._write_if_exists(k, data, signed=True)

        for k in (
            0x3D,
            0x3E,
            0x3F,
            0x40,
            0x41,
            0x42,
            0x43,
            0x44,
            0x45,
            0x46,
            0x47,
            0x48,
            0x49,
            0x4A,
            0x4B,
        ):
            self._write_if_exists(k, data)

        # PANOSE
        values = data[ttinfo_names[0x4C]]
        assert len(values) == 10
        self.write_uint8(0x4C)
        for value in values:
            self.write_uint8(value)

        for k in (0x4D, 0x4E, 0x4F, 0x50, 0x51, 0x52, 0x5C):
            self._write_if_exists(k, data)

        # HDMX 1 and 2
        for k in (0x53, 0x58):
            strkey = ttinfo_names[k]
            if strkey in data:
                values = data[strkey]
                self.write_uint8(k)
                self.write_value(len(values))
                for value in values:
                    self.write_uint8(value)

        # Codepages
        cp_dict = data[ttinfo_names[0x54]]
        self.write_uint8(0x54)
        for ck in ("os2_ul_code_page_range1", "os2_ul_code_page_range2"):
            value = cp_dict[ck]
            self.write_value(value, signed=False)

        self.write_uint8(0x32)  # End marker


class TrueTypeStemPpems1Compiler(BaseCompiler):
    def _compile(self, data: "TTStemsDict") -> None:
        for direction in ("ttStemsV", "ttStemsH"):
            stems = data[direction]
            for stem in stems:
                self.write_value(stem["round"][1])


class TrueTypeStemPpems23Compiler(BaseCompiler):
    def _compile(self, data: "TTStemsDict") -> None:
        for direction in ("ttStemsV", "ttStemsH"):
            stems = data[direction]
            self.write_value(len(stems))
            for stem in stems:
                for k in (2, 3):
                    self.write_value(stem["round"][k])


class TrueTypeStemPpemsCompiler(BaseCompiler):
    def _compile(self, data: "TTStemsDict") -> None:
        for direction in ("ttStemsV", "ttStemsH"):
            stems = data[direction]
            self.write_value(len(stems))
            for stem in stems:
                for k in range(2, 6):
                    self.write_value(stem["round"][k])


class TrueTypeStemsCompiler(BaseCompiler):
    def _compile(self, data: "TTStemsDict") -> None:
        for direction in ("ttStemsV", "ttStemsH"):
            stems = data[direction]
            self.write_value(len(stems))
            for stem in stems:
                self.write_value(stem["value"])  # width
                stem_name = stem["name"].encode(self.encoding)
                self.write_uint8(len(stem_name))
                self.write_bytes(stem_name)
                self.write_value(stem["round"][6])


class TrueTypeZonesCompiler(BaseCompiler):
    """
    A compiler that compiles TrueType hinting "alignment zones" data
    """

    def _compile(self, data: "TTZonesDict") -> None:
        for side in ("ttZonesT", "ttZonesB"):
            if side not in data:
                self.write_value(0)
                continue

            side_zones: "list[TTZoneDict]" = data[side]
            self.write_value(len(side_zones))
            for zone in side_zones:
                self.write_value(zone["position"])
                self.write_value(zone["value"])
                name = zone["name"]
                self.write_str_with_len(name)


class TrueTypeZoneDeltasCompiler(BaseCompiler):
    """
    A compiler that compiles TrueType hinting "alignment zones" deltas data
    """

    def _compile(self, data: dict[int, dict[int, int]]) -> None:
        deltas = []
        for zone_index, spec in data.items():
            for ppm, shift in spec.items():
                deltas.append((zone_index, ppm, shift))
        deltas.sort()
        self.write_value(len(deltas))
        for zone_index, ppm, shift in deltas:
            self.write_value(zone_index)
            self.write_value(ppm)
            self.write_value(shift)


class VdmxCompiler(BaseCompiler):
    def _compile(self, data: "list[VdmxRecDict]") -> None:
        self.write_value(len(data))
        for rec in data:
            self.write_value(rec["pelHeight"])
            self.write_value(rec["max"])
            self.write_value(rec["min"])



from struct import unpack

from vfbLib.helpers import binaryToIntList
from vfbLib.parsers.base import BaseParser


class DoubleParser(BaseParser):
    """
    A parser that reads data as a double-size float.
    """

    def _parse(self):
        return self.read_double()


class FloatListParser(BaseParser):
    """
    A parser that reads data as a list of floats.
    """

    __size__ = 4
    __fmt__ = "f"

    def _parse(self):
        values = []
        for _ in range(self.stream.getbuffer().nbytes // self.__size__):
            values.extend(unpack(self.__fmt__, self.stream.read(self.__size__)))

        return values


class DoubleListParser(FloatListParser):
    """
    A parser that reads data as a list of doubles.
    """

    __size__ = 8
    __fmt__ = "d"


class Int16Parser(BaseParser):
    """
    A parser that reads data as UInt16.
    """

    def _parse(self):
        return self.read_uint16()


# class Int32Parser(BaseParser):
#     """
#     A parser that reads data as UInt32.
#     """

#     def _parse(self):
#         return self.read_uint32()


# class Int64Parser(BaseParser):
#     """
#     A parser that reads data as UInt64.
#     """

#     def _parse(self):
#         return int.from_bytes(self.stream.read(64), byteorder="little", signed=False)


class IntListParser(BaseParser):
    """
    A parser that reads data as a list of UInt16.
    """

    __size__ = 4

    def _parse(self):
        values = []
        for _ in range(self.stream.getbuffer().nbytes // self.__size__):
            values.append(
                int.from_bytes(
                    self.stream.read(self.__size__),
                    byteorder="little",
                    signed=False,
                )
            )
        return values


class PanoseParser(BaseParser):
    """
    A parser that reads data as an array representing PANOSE values.
    """

    def _parse(self):
        return list(unpack("<10b", self.stream.read()))


class SignedInt16Parser(BaseParser):
    """
    A parser that reads data as signed int16.
    """

    def _parse(self):
        return self.read_int16()


class SignedInt32Parser(BaseParser):
    """
    A parser that reads data as signed int32.
    """

    def _parse(self):
        # return int.from_bytes(self.stream.read(), byteorder="little", signed=True)
        return self.read_int32()


class UnicodeRangesParser(BaseParser):
    """
    A parser that reads data as uint64 and returns it as a list of bit numbers.
    """

    def _parse(self):
        result = int.from_bytes(self.stream.read(64), byteorder="little", signed=False)
        return binaryToIntList(result)



from struct import pack
from typing import Any

from vfbLib.compilers.base import BaseCompiler


class DoubleCompiler(BaseCompiler):
    """
    A compiler that compiles double-precision float data.
    """

    def _compile(self, data: Any) -> None:
        self.write_double(data)


class DoubleListCompiler(BaseCompiler):
    """
    A compiler that compiles a list of doubles.
    """

    def _compile(self, data: Any) -> None:
        self.write_doubles(data)


class Int16Compiler(BaseCompiler):
    """
    A compiler that compiles UInt16 data.
    """

    def _compile(self, data: Any) -> None:
        self.write_uint16(data)


class IntListCompiler(BaseCompiler):
    """
    A compiler that compiles a list of ints.
    """

    __size__ = 4

    def _compile(self, data: Any) -> None:
        for value in data:
            self.write_bytes(
                value.to_bytes(self.__size__, byteorder="little", signed=False)
            )


class PanoseCompiler(BaseCompiler):
    """
    A compiler that compiles PANOSE data.
    """

    def _compile(self, data: Any) -> None:
        b = pack("<10b", *data)
        self.write_bytes(b)


class SignedInt16Compiler(BaseCompiler):
    """
    A compiler that compiles Int16 data.
    """

    def _compile(self, data: Any) -> None:
        self.write_int16(data)


class SignedInt32Compiler(BaseCompiler):
    """
    A compiler that compiles Int32 data.
    """

    def _compile(self, data: Any) -> None:
        self.write_int32(data)


class UnicodeRangesCompiler(BaseCompiler):
    """
    A compiler that compiles the unicoderanges value into an uint64.
    """

    def _compile(self, data: list[int]) -> None:
        value = 0
        for b in data:
            value += 2**b
        self.write_bytes(value.to_bytes(16, byteorder="little", signed=False))



import logging
from enum import Enum
from struct import unpack
from typing import TYPE_CHECKING, Any

from fontTools.ttLib.tables.ttProgram import Program

from vfbLib import DIRECTIONS, gdef_class_names, replace_types
from vfbLib.helpers import hexStr
from vfbLib.parsers.base import BaseParser
from vfbLib.parsers.guides import parse_guides
from vfbLib.truetype import TT_COMMANDS
from vfbLib.typing import (
    AnchorDict,
    AnchorPropertiesDict,
    Component,
    GdefDict,
    GlyphData,
    HintDict,
    Instruction,
    LinkDict,
    MaskData,
    MMAnchorDict,
    MMHintsDict,
    MMNode,
    Point,
)
from vfbLib.value import read_value

if TYPE_CHECKING:
    from io import BytesIO


logger = logging.getLogger(__name__)


class PathCommand(Enum):
    move = 0
    line = 1
    curve = 3
    qcurve = 4


def read_absolute_point(
    points: list[list[Any]],
    stream: "BytesIO",
    num_masters: int,
    x_masters: list[int],
    y_masters: list[int],
) -> None:
    # Read coordinates for all masters from stream, add the points to a list
    # of points per master. Keep track of the relative coordinates reference
    # value in x_masters and y_masters.
    for m in range(num_masters):
        # End point
        xrel = read_value(stream)
        yrel = read_value(stream)
        x_masters[m] += xrel
        y_masters[m] += yrel
        points[m].append((x_masters[m], y_masters[m]))


class GlyphAnchorsParser(BaseParser):
    def _parse(self) -> list[MMAnchorDict]:
        anchors = []
        num_anchors = self.read_value()
        num_masters = self.read_value()
        for _ in range(num_anchors):
            anchor = MMAnchorDict(x=[], y=[])
            for _ in range(num_masters):
                anchor["x"].append(self.read_value())
                anchor["y"].append(self.read_value())
            anchors.append(anchor)
        return anchors


class GlyphAnchorsSuppParser(BaseParser):
    def _parse(self) -> list[AnchorPropertiesDict]:
        anchors = []
        num_anchors = self.read_value()
        for _ in range(num_anchors):
            hue = self.read_value()
            data = self.read_value()
            anchors.append(AnchorPropertiesDict(hue=hue, data=data))
        return anchors


class GlyphGDEFParser(BaseParser):
    def _parse(self) -> GdefDict:
        gdef: GdefDict = {}
        glyph_class = self.read_value()
        glyph_class_name = gdef_class_names[glyph_class]
        if glyph_class_name != "unassigned":
            gdef["glyph_class"] = glyph_class_name

        num_anchors = self.read_value()
        anchors = []
        for _ in range(num_anchors):
            name = self.read_str_with_len()
            x = self.read_value()
            _ = self.read_value()
            y = self.read_value()
            _ = self.read_value()
            anchor = AnchorDict(x=x, y=y)
            if name:
                anchor["name"] = name
            anchors.append(anchor)
        if anchors:
            gdef["anchors"] = anchors

        num_carets = self.read_value()
        carets = []
        for _ in range(num_carets):
            pos = self.read_value()
            idx = self.read_value()
            carets.append((pos, idx))
        if carets:
            gdef["carets"] = carets

        num_ot_classes = self.read_value()
        if num_ot_classes > 0:
            ot_classes = [self.read_value() for _ in range(num_ot_classes)]
            gdef["ot_classes"] = ot_classes

        return gdef


class GlyphOriginParser(BaseParser):
    def _parse(self) -> dict[str, Any]:
        x = self.read_int16()
        y = self.read_int16()
        return {"x": x, "y": y}


class GlyphParser(BaseParser):
    def _parse(self) -> dict[str, Any]:
        """
        01090701
        01  92[7]2e 6e 6f 74 64 65 66 # Glyph name
        08  8c bb[48] 93[8]
            00 ab 8b    [  32,    0] move
            01 8b f9a0  [   0,  780] line 32,780
            01 f7fc 8b  [ 360,    0] line 392,780
            01 8b fda0  [   0, -780] line 392, 0
            00 fbcc bb  [-312,   48] move  80, 48
            01 f79c 8b  [ 264,    0] line
            01 8b f940  [   0,  684] line
            01 fb9c 8b  [-264,    0] line
        02  f83c 8b  [ 424,    0]
        03  8b 8b 8b
        04  8b 8b
        0a  a1[22] 8f[4] # TT hinting
            02 8b 8b
            04 8b 8c 89 8a
            04 8b 8f 89 8a
            04 8c 92 89 8a
            8b 8b 8b
        0f
        """
        self.glyphdata = GlyphData()
        start = unpack("<4B", self.stream.read(4))
        if start != (1, 9, 7, 1):
            logger.warning(f"Unexpected glyph constant: {start}")
        while True:
            # Read a value to decide what kind of information follows
            v = self.read_uint8()

            match v:
                case 0x01:
                    # Glyph name
                    glyph_name = self.read_str_with_len()
                    self.glyphdata["name"] = glyph_name
                    self.name = self.glyphdata["name"]

                case 0x02:
                    # Metrics
                    self.parse_metrics()

                case 0x03:
                    # PS Hints
                    self.parse_hints()

                case 0x04:
                    # Guides
                    self.parse_guides()

                case 0x05:
                    # Components
                    self.parse_components()

                case 0x06:
                    # Kerning
                    self.parse_kerning()

                case 0x07:
                    # Image
                    logger.warning("Skipping glyph data: 0x07 (image), not implemented")

                case 0x08:
                    # Outlines
                    self.num_masters = self.parse_outlines(self.glyphdata)

                case 0x09:
                    # Imported binary TrueType data
                    self.parse_binary()

                case 0x0A:
                    # TrueType instructions
                    self.parse_instructions()

                case 0x0F:
                    # End of glyph
                    break

                case _:
                    logger.error(f"Unhandled info field: {hex(v)}")
                    logger.error(hexStr(self.stream.read()))
                    raise ValueError

        return dict(self.glyphdata)

    def parse_guides(self) -> None:
        # Guidelines
        guides = parse_guides(self.stream, self.num_masters, f"glyph '{self.name}'")
        if guides:
            self.glyphdata["guides"] = guides

    def parse_binary(self) -> None:
        # Imported binary TrueType data
        imported: dict[str, Any] = {}
        while True:
            key = self.read_uint8()

            match key:
                case 0x28:
                    # End
                    break

                case 0x29:
                    # Metrics
                    imported["width"] = self.read_value(signed=False)
                    imported["lsb"] = self.read_value()
                    imported["unknown1"] = self.read_value()
                    imported["unknown2"] = self.read_value()
                    imported["unknown3"] = self.read_value()
                    imported["bbox"] = [self.read_value() for _ in range(4)]

                case 0x2A:
                    # Outlines
                    num_contours = self.read_value()  # -1 for composite
                    imported["num_contours"] = num_contours
                    imported["endpoints"] = [
                        self.read_value(signed=False) for _ in range(num_contours)
                    ]
                    num_nodes = self.read_value(signed=False)
                    nodes = []
                    x = 0
                    y = 0
                    for _ in range(num_nodes):
                        x += self.read_value()
                        y += self.read_value()
                        flags = self.read_uint8()
                        # flags = byte >> 4
                        # cmd = byte & 0x0F
                        # bin_flags = f"{flags:08b}"
                        node = {
                            "flags": flags,
                            "on": int(bool(flags & 1)),
                            # "x_short": int(bool(flags & 2)),
                            # "y_short": int(bool(flags & 4)),
                            # "repeat_flag": int(bool(flags & 8)),
                            # "x_same": int(bool(flags & 16)),
                            # "y_same": int(bool(flags & 32)),
                            # "overlap": int(bool(flags & 64)),
                            "point": (x, y),
                        }
                        nodes.append(node)
                    if nodes:
                        imported["nodes"] = nodes

                case 0x2B:
                    # Instructions
                    num_bytes = self.read_value(signed=False)
                    instructions = self.stream.read(num_bytes)
                    p = Program()
                    p.fromBytecode(instructions)
                    imported["instructions"] = p.getAssembly()

                case 0x2C:
                    # HDMX data
                    num = self.read_value()
                    imported["hdmx"] = [self.read_uint8() for _ in range(num)]

                case _:
                    logger.error(imported)
                    logger.error(
                        f"Unknown key in imported binary glyph data: {hex(key)}"
                    )
                    raise ValueError

        self.glyphdata["imported"] = imported

    def parse_components(self) -> None:
        components = []
        num = self.read_value()
        for _ in range(num):
            gid = self.read_value()
            c = Component(gid=gid, offsetX=[], offsetY=[], scaleX=[], scaleY=[])
            for _ in range(self.num_masters):
                x = self.read_value()
                y = self.read_value()
                scaleX, scaleY = self.read_doubles(2)
                c["offsetX"].append(x)
                c["offsetY"].append(y)
                c["scaleX"].append(scaleX)
                c["scaleY"].append(scaleY)
            components.append(c)
        self.glyphdata["components"] = components

    def parse_hints(self) -> None:
        hints = MMHintsDict(v=[], h=[])
        for direction in DIRECTIONS:
            num_hints = self.read_value()
            for _ in range(num_hints):
                master_hints = []
                for _ in range(self.num_masters):
                    pos = self.read_value()
                    width = self.read_value()
                    master_hints.append(HintDict(pos=pos, width=width))
                hints[direction].append(master_hints)

        num_hintmasks = self.read_value()
        if num_hintmasks > 0:
            hintmasks: list[tuple[str, int]] = []
            for _ in range(num_hintmasks):
                k = self.read_uint8()
                val = self.read_value()
                key = replace_types[k]
                hintmasks.append((key, val))
            if hintmasks:
                hints["hintmasks"] = hintmasks

        if hints["v"] or hints["h"] or "hintmasks" in hints:
            self.glyphdata["hints"] = hints

    def parse_instructions(self) -> None:
        # Number of bytes for instructions that follow;
        # we don't use it
        _ = self.read_value()
        num_commands = self.read_value()
        commands: list[Instruction] = []
        for i in range(num_commands):
            cmd = self.read_uint8()
            params = [self.read_value() for _ in range(len(TT_COMMANDS[cmd]["params"]))]
            commands.append(
                Instruction(
                    cmd=TT_COMMANDS[cmd]["name"],
                    params=dict(zip(TT_COMMANDS[cmd]["params"], params)),
                )
            )
        # Instructions are ended by 3 * 0!?
        for _ in range(3):
            self.read_value()

        if commands:
            self.glyphdata["tth"] = commands

    def parse_metrics(self) -> None:
        metrics: list[Point] = []
        for _ in range(self.num_masters):
            master_metrics = (
                self.read_value(),
                self.read_value(),
            )
            metrics.append(master_metrics)
        self.glyphdata["metrics"] = metrics

    def parse_outlines(self, target: GlyphData | MaskData) -> int:
        # Nodes
        num_masters = self.read_value(signed=False)
        target["num_masters"] = num_masters

        # 2 x the number of values to be read after num_nodes, the reason is unclear.
        _ = self.read_value()
        # glyphdata["num_node_values"] = num_node_values

        num_nodes = self.read_value()
        segments: list[MMNode] = []
        x = [0 for _ in range(num_masters)]
        y = [0 for _ in range(num_masters)]

        for _ in range(num_nodes):
            byte = self.read_uint8()
            flags = byte >> 4
            cmd = PathCommand(byte & 0x0F).name

            # End point
            points: list[list[Point]] = [[] for _ in range(num_masters)]
            read_absolute_point(points, self.stream, num_masters, x, y)

            if cmd == "curve":
                # First control point
                read_absolute_point(points, self.stream, num_masters, x, y)
                # Second control point
                read_absolute_point(points, self.stream, num_masters, x, y)

            segment = MMNode(type=cmd, flags=flags, points=points)
            segments.append(segment)
        target["nodes"] = segments
        return num_masters

    def parse_kerning(self) -> None:
        num = self.read_value()
        kerning = {}
        for _ in range(num):
            # Glyph index of right kerning partner
            gid = self.read_value()
            values = []
            for _ in range(self.num_masters):
                # List of values, one value per master
                values.append(self.read_value())
            kerning[gid] = values
        self.glyphdata["kerning"] = kerning


class GlyphSketchParser(BaseParser):
    def _parse(self) -> list[tuple[int, int, int]]:
        num = self.read_value(signed=False)
        return [
            (self.read_value(), self.read_value(), self.read_value())
            for _ in range(num)
        ]


class GlyphUnicodeParser(BaseParser):
    def _parse(self) -> list[int]:
        unicodes = []
        for _ in range(self.stream.getbuffer().nbytes // 2):
            u = self.read_uint16()
            unicodes.append(u)
        return unicodes


class GlyphUnicodeSuppParser(BaseParser):
    def _parse(self) -> list[int]:
        unicodes = []
        for _ in range(self.stream.getbuffer().nbytes // 4):
            u = self.read_uint32()
            unicodes.append(u)
        return unicodes


class LinkParser(BaseParser):
    def _parse(self) -> LinkDict:
        links = LinkDict(x=[], y=[])
        for i in range(2):
            num = self.read_value()
            for _ in range(num):
                src = self.read_value()
                tgt = self.read_value()
                links[("y", "x")[i]].append((src, tgt))
        return links


class MaskParser(GlyphParser):
    def _parse(self) -> dict[str, Any]:
        num_values = self.read_value()
        weight_vector = []
        for _ in range(num_values):
            weight_vector.append(self.read_value() / 100_000_000)
        maskdata = MaskData(weight_vector=weight_vector)

        # From here, the mask is equal to the outlines
        self.parse_outlines(maskdata)
        return dict(maskdata)


class GlobalMaskParser(GlyphParser):
    def _parse(self) -> dict[str, Any]:
        maskdata = GlyphData()
        self.parse_outlines(maskdata)
        return dict(maskdata)


class MaskMetricsParser(BaseParser):
    # advance width/height for master 0
    def _parse(self) -> tuple[int, int]:
        # FIXME: Normally, the values are little endian.
        # In some cases, the value seems to be saved as big endian, though
        # Maybe try to interpret the value as BE when it is negative and less than a
        # certain sensible value.
        x = self.read_int16()
        y = self.read_int16()
        return (x, y)


class MaskMetricsMMParser(BaseParser):
    # advance width/height for master 1 to 15
    def _parse(self) -> list[tuple[int, int]]:
        assert self.vfb is not None
        values = []
        for _ in range(self.vfb.num_masters - 1):
            x = self.read_value()
            y = self.read_value()
            values.append((x, y))
        return values



import logging
from io import BytesIO
from struct import unpack
from typing import TYPE_CHECKING, Any

from vfbLib import mapping_modes
from vfbLib.helpers import (
    deHexStr,
    double_size,
    hexStr,
    int8_size,
    int16_size,
    int32_size,
)
from vfbLib.parsers.value import read_value
from vfbLib.typing import MappingModeDict

if TYPE_CHECKING:
    from io import BufferedIOBase

    from vfbLib.typing import KerningClassFlagDict, MetricsClassFlagDict
    from vfbLib.vfb.vfb import Vfb


logger = logging.getLogger(__name__)


class StreamReader:
    """
    Base class that reads values from the input stream.

    This is the parent class for the general BaseParser, from which all other parsers
    inherit, but it may be subclassed directly if more flexibility is needed.
    """

    def __init__(self) -> None:
        self.encoding = "cp1252"
        self.stream: "BufferedIOBase" = BytesIO()

    def read_double(self) -> float:
        """
        Return a double-precision float from the stream.

        Returns:
            float: The float
        """
        return self.read_doubles(1)[0]

    def read_doubles(self, num) -> tuple[float, ...]:
        """
        Return a tuple of `num` double-precision floats from the stream.

        Args:
            num (int): The number of double-precision floats to read from the stream

        Returns:
            tuple[float]: The tuple of floats
        """
        return unpack(num * "d", self.stream.read(num * double_size))

    def read_int8(self) -> int:
        """
        Return a signed 8-bit integer from the stream.

        Returns:
            int: The integer
        """
        return int.from_bytes(
            self.stream.read(int8_size), byteorder="little", signed=True
        )

    def read_int16(self) -> int:
        """
        Return a signed 16-bit integer from the stream.

        Returns:
            int: The integer
        """
        return int.from_bytes(
            self.stream.read(int16_size), byteorder="little", signed=True
        )

    def read_int32(self) -> int:
        """
        Return a signed 32-bit integer from the stream.

        Returns:
            int: The integer
        """
        return int.from_bytes(
            self.stream.read(int32_size), byteorder="little", signed=True
        )

    def read_str(self, size: int) -> str:
        """
        Return a string of the specified `size` from the current stream with the current
        encoding.

        Args:
            size (int): The size in bytes to be converted to a string

        Returns:
            str: The string
        """
        if size == 0:
            return ""
        return self.stream.read(size).decode(self.encoding)

    def read_str_with_len(self) -> str:
        """
        Read the length of a string from the current stream, then the string itself, and
        return the string.

        Returns:
            str: The string
        """
        size = self.read_value(signed=False)
        return self.read_str(size)

    def read_str_all(self) -> str:
        """
        Return the remaining bytes of the current stream as a string with the current
        encoding.

        Returns:
            str: The string
        """
        return self.stream.read().decode(self.encoding)

    def read_uint8(self) -> int:
        """
        Return an unsigned 8-bit integer from the stream.

        Returns:
            int: The integer
        """
        return int.from_bytes(
            self.stream.read(int8_size), byteorder="little", signed=False
        )

    def read_uint16(self) -> int:
        """
        Return an unsigned 16-bit integer from the stream.

        Returns:
            int: The integer
        """
        return int.from_bytes(
            self.stream.read(int16_size), byteorder="little", signed=False
        )

    def read_uint32(self) -> int:
        """
        Return an unsigned 32-bit integer from the stream.

        Returns:
            int: The integer
        """
        return int.from_bytes(
            self.stream.read(int32_size), byteorder="little", signed=False
        )

    def read_value(self, signed: bool = True) -> int:
        """
        Return an encoded integer value from the stream.

        Args:
            signed (bool, optional): Whether the value is interpreted as a signed value.
            Defaults to True.

        Returns:
            int: The integer
        """
        return read_value(self.stream, signed=signed)


class BaseParser(StreamReader):
    """
    Base class to parse data from a vfb file
    """

    def __init__(self) -> None:
        self.encoding = "cp1252"
        self.stream: BytesIO = BytesIO()

    def parse(self, stream: BytesIO, size: int, vfb: "Vfb | None") -> Any:
        """
        Prepare the parsing of the stream, then call the specialized parser and return
        the decompiled VFB entry structure.

        The specialized parsing is done by calling the `_parse` method, which must be
        implemented for all entry parser sublasses.

        Args:
            stream (BytesIO): The stream to read from.
            size (int): The number of bytes that will be read from the input stream and
                parsed.
            vfb (int, optional): The Vfb that is calling the parser.

        Raises:
            AssertionError: If bytes remain in the stream after the parsing finished.

        Returns:
            Any: The parsed structure. The type depends on the specific entry that is
            being parsed.
        """
        self.stream = BytesIO(stream.read(size))  # type: ignore
        self.vfb = vfb
        decompiled = self._parse()

        # Make sure the parser consumed all of the data
        remainder = self.stream.read()
        if remainder != b"":
            logger.error(f"Parser {self.__class__.__name__} did not consume all bytes.")
            logger.error(f"Remainder: {remainder!r}")
            raise AssertionError

        return decompiled

    def parse_hex(self, hexstr: str, vfb: "Vfb | None" = None):
        """
        Parse the data given in hex string format, e.g. "8c 8d 89 8b". Used for testing.

        Args:
            hexstr (str): The data
        """
        data = deHexStr(hexstr)
        return self.parse(BytesIO(data), len(data), vfb)

    def _parse(self) -> Any:
        """
        Custom parsing method. By default, it returns a human-readable hex string of the
        data.

        Returns:
            Any: _description_
        """
        return hexStr(self.stream.read())


class EncodedKeyValuesParser(BaseParser):
    __end__ = 0x64

    def _parse(self) -> Any:
        """
        Parse and return the entry as a list of key-value-dictionaries. Both keys and
        values in the dictionary are integers.

        Returns:
            list[dict[int, int]]: The list of key-value dictionaries.
        """
        values: list[dict[int, int]] = []
        while True:
            key = self.read_uint8()
            if key == self.__end__:
                return values

            val = self.read_value()
            values.append({key: val})


class MappingModeParser(BaseParser):
    __end__ = 0x00

    def _parse(self) -> MappingModeDict:
        while True:
            key = self.read_uint8()
            if key == self.__end__:
                break

            v = self.read_value()

            if key == 1:
                mapping_mode = mapping_modes.get(v, str(v))
            elif key == 2:
                m2 = v
            elif key == 3:
                m3 = v
            elif key == 4:
                # encoding id:
                #   in mapping_mode 0:
                #      -1: glyph index mode
                #       0: imported encoding
                #       n: encoding id
                #   in mapping_mode 1:
                #      index into unicode_ranges from uranges.dat
                #   in mapping mode 3:
                #      index into codepages sorted by group, then name (?)
                mapping_id = v
            else:
                raise KeyError

        return MappingModeDict(
            mapping_mode=mapping_mode, m2=m2, m3=m3, mapping_id=mapping_id
        )


class EncodedValueListParser(BaseParser):
    """
    A parser that reads data as Yuri's optimized encoded values.
    """

    def _parse(self) -> list[int]:
        values = []
        while True:
            try:
                val = self.read_value()
                values.append(val)
            except EOFError:
                logger.debug("EOF")
                return values


class EncodedValueListWithCountParser(BaseParser):
    """
    A parser that reads data as Yuri's optimized encoded values. The list of values is
    preceded by a count value that specifies how many values should be read.
    """

    def _parse(self) -> list[int]:
        count = self.read_value()
        values: list[int] = []  # TODO: We don't need the dict
        for _ in range(count):
            val = self.read_value()
            values.append(val)
        return values


class GlyphEncodingParser(BaseParser):
    def _parse(self) -> tuple[int, str]:
        gid = self.read_uint16()
        nam = self.stream.read().decode("cp1252")
        return gid, nam


class OpenTypeKerningClassFlagsParser(BaseParser):
    def _parse(self) -> "KerningClassFlagDict":
        class_flags: "KerningClassFlagDict" = {}
        num_classes = self.read_value()
        for _ in range(num_classes):
            name = self.read_str_with_len()
            flag1 = self.read_value()
            flag2 = self.read_value()
            class_flags[name] = (flag1, flag2)
        return class_flags


class OpenTypeMetricsClassFlagsParser(BaseParser):
    def _parse(self) -> "MetricsClassFlagDict":
        class_flags: "MetricsClassFlagDict" = {}
        num_classes = self.read_value()
        for _ in range(num_classes):
            name = self.read_str_with_len()
            flag1 = self.read_value()
            flag2 = self.read_value()
            flag3 = self.read_value()
            class_flags[name] = (flag1, flag2, flag3)
        return class_flags



from vfbLib import ExpandKernOptions, TTAutoHintOptions, export_options, font_options
from vfbLib.helpers import binaryToIntList
from vfbLib.parsers.base import BaseParser, EncodedKeyValuesParser
from vfbLib.typing import ExpandKernFlagsDict, FontOptionsDict, TTAutoHintOptionsDict


class ExportOptionsParser(BaseParser):
    """
    A parser that reads data a bit field representing export options.
    """

    def _parse(self) -> list[str | int]:
        val = self.read_uint16()
        bits = binaryToIntList(val)
        options = [export_options.get(i, i) for i in bits]
        return options


class FontOptionsParser(EncodedKeyValuesParser):
    def _parse(self) -> FontOptionsDict:
        options = FontOptionsDict()
        for d in super()._parse():
            assert len(d) == 1
            k, v = tuple(d.items())[0]
            options[font_options.get(k, str(k))] = v

        # Post-process some entries

        for key in ("auto_hinting_h_ratio", "auto_hinting_v_ratio"):
            if key in options:
                options[key] /= 10_000

        if "autohinting_options" in options:
            val = options["autohinting_options"]
            if isinstance(val, int):
                # Convert to dict
                options["autohinting_options"] = TTAutoHintOptionsDict(
                    single_link_attachment_precision=val
                    & int(TTAutoHintOptions.single_link_attachment_precision),
                    generate_triple_hints=int(
                        bool(val & TTAutoHintOptions.generate_triple_hints)
                    ),
                    generate_delta_instructions=int(
                        bool(val & TTAutoHintOptions.generate_delta_instructions)
                    ),
                    direct_links_to_center_of_the_glyph_where_possible=int(
                        bool(
                            val
                            & TTAutoHintOptions.direct_links_to_center_of_the_glyph_where_possible  # noqa: E501
                        )
                    ),
                    interpolate_positions_of_cusp_points=int(
                        bool(
                            val & TTAutoHintOptions.interpolate_positions_of_cusp_points
                        )
                    ),
                    interpolate_positions_of_double_links=int(
                        bool(
                            val
                            & TTAutoHintOptions.interpolate_positions_of_double_links
                        )
                    ),
                    add_link_to_rsb=int(bool(val & TTAutoHintOptions.add_link_to_rsb)),
                )

        if "expand_kern_flags" in options:
            val2 = options["expand_kern_flags"]
            if isinstance(val2, int):
                options["expand_kern_flags"] = ExpandKernFlagsDict(
                    limit_action=int(bool(val2 & ExpandKernOptions.limit_action)),
                    limit_codepage=int(bool(val2 & ExpandKernOptions.limit_codepage)),
                    limit_cmap_10=int(bool(val2 & ExpandKernOptions.limit_cmap_10)),
                    limit_font_window=int(
                        bool(val2 & ExpandKernOptions.limit_font_window)
                    ),
                    limit_count=int(bool(val2 & ExpandKernOptions.limit_count)),
                    limit_keep=int(bool(val2 & ExpandKernOptions.limit_keep)),
                    apply_to_assistance=int(
                        bool(val2 & ExpandKernOptions.apply_to_assistance)
                    ),
                )

        return options



import logging
from typing import TYPE_CHECKING

from vfbLib.compilers.base import BaseCompiler
from vfbLib.parsers.flversion import END, OWNER, PLATFORM, VERSION, platform_ids

if TYPE_CHECKING:
    from vfbLib.typing import FLVersionDict

logger = logging.getLogger(__name__)


class FLVersionCompiler(BaseCompiler):
    def _compile(self, data: "FLVersionDict") -> None:
        self.write_uint8(PLATFORM)
        self.write_value(platform_ids[data["platform"]])

        self.write_uint8(VERSION)
        version_tuple = data["version"]
        assert len(version_tuple) == 4
        self.write_uint8(0xFF)
        for value in version_tuple:
            self.write_uint8(value)

        self.write_uint8(OWNER)
        self.write_value(data["owner"])

        self.write_uint8(END)



from re import search


def rename_kern_classes_in_feature_code(data: list[str]) -> list[str]:
    """
    Convert references to kerning classes in the feature code to current UFO naming
    standards (public.kern1, public.kern2).
    """
    features = []
    for line in data:
        if "#" in line:
            code, comment = line.split("#", 1)
        else:
            code = line
            comment = None

        if m := search(r"^([\t\s]*)(enum )?(pos) +([\S]+) ([\S]+) +([\S]+.*)$", code):
            indent, enum, pos, L, R, value = m.groups()
            if L.startswith("@_"):
                L = f"@public.kern1.{L[1:]}"
            if R.startswith("@_"):
                R = f"@public.kern2.{R[1:]}"
            if enum is None:
                enum = ""
            code = f"{indent}{enum}{pos} {L} {R} {value}"
            if comment:
                code += f"#{comment}"
            features.append(code)
        else:
            features.append(line)

    return features



from pathlib import Path

import orjson

from vfbLib.vfb.vfb import Vfb


def save_vfb_json(
    vfb_path: Path,
    out_path: Path | None = None,
    only_header: bool = False,
    minimal: bool = False,
    unicode_strings: bool = False,
    no_decompile: bool = False,
    roundtrip: bool = False,
) -> None:
    vfb = Vfb(
        vfb_path,
        only_header=only_header,
        minimal=minimal,
        unicode_strings=unicode_strings,
    )
    if roundtrip:
        vfb.decompile()
        vfb.compile()

    if not no_decompile:
        vfb.decompile()

    suffix = ".vfb.json"
    if out_path:
        out_path = (Path(out_path) / vfb_path.name).with_suffix(suffix)
    else:
        out_path = vfb_path.with_suffix(suffix)
    write_vfb_json(vfb, out_path)


def write_vfb_json(vfb: Vfb, out_path: Path) -> None:
    with open(str(out_path), "wb") as f:
        f.write(
            orjson.dumps(
                vfb.as_dict(), option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS
            )
        )
